# Aiko Engine
# ~~~~~~~~~~~
#
# Usage
# ~~~~~
# import time
# import aiko_services.event as event
#
# counter = 1
# def flatout_test():
#     global counter
#     counter += 1
#
# def timer_test():
#     global counter
#     print(f"timer_test(): {time.monotonic()}: {counter}")
#
# event.add_flatout_handler(flatout_test)
# event.add_timer_handler(timer_test, 1.0)
# event.loop()
#
# To Do
# ~~~~~
# * Remove queue handler ... and replace usage with mailboxes *
#
# * For shared data, e.g Locks and more ...
#   - Ensure all updates occur via "events" (handlers, messages and timers)
#     managed by the event loop and executed solely by the event loop thread
#
# * Replace 'event.queue_put(message, "message")' with the general ability to
#   put any 'object' onto a queue to get "work" onto the "event loop" and
#   have that work done by a specified handler
#
# * Rename "event.py" to "handler.py" along with function and variable names ?
#
# * Make function names consistently verb_noun() or noun_verb(), but not both !
#
# * Improve event.loop() to check timer in-between every mailbox check
#
# - Provide unit tests !  For all add / remove all the various handler types.
#
# - BUG: With multiple timer handlers using the same handler function,
#        then remove_timer_handler() make remove the wrong one.
#        Need an identifier to specific exactly which handler instance
#
# - BUG: Handle case of calling event.terminate() before entering event.loop()
#        Due to event_enabled being overwritten at the start of event.loop()
#
# - BUG: Make _handler_count thread-safe, when adding and removing handlers !
#   - See "mutex": https://hg.python.org/cpython/file/3.5/Lib/queue.py
#
# - BUG: Why isn't event.add_timer_handler(_, _, True) firing immediately ?
#
# - add_timer_handler(... count=None) to automatically expire after count
# - add_mailbox_handler(... max_size=None) to limit maximum queue size
#
# - Move "_timer_counter" and "update_timer_counter()" into EventList
# - Refactor all remaining functions into new EventEngine class
# - Coalesce remove_flatout_handler() and remove_timer_handler() into
#     single remove_handler() that works in all cases.
#
# - Provide simple wallclock time monitoring of handler invocation time
# - Currently, flatout_handlers invocation rate is limited by ...
#     "sleep_time = 0.001", i.e 1,000 Hz.  Make this configurable.
# - All Services should have initialise() and stream event handler()
#   - All Streams also have task_start() and task_stop()
# - Since handlers take time, need to adjust time.sleep() period
# - New event types: GStreamer appsink, appsrc, serial

from collections import OrderedDict  # All OrderedDict operations are O(1)
import queue
import time
from typing import Any, Tuple

from aiko_services.main.utilities import *

__all__ = [
    "add_flatout_handler", "add_mailbox_handler",
    "add_queue_handler", "add_timer_handler",
    "loop", "mailbox_put", "queue_put",
    "remove_flatout_handler", "remove_mailbox_handler",
    "remove_queue_handler", "remove_timer_handler",
    "terminate"
]

_MAILBOX_INCREMENT_WARNING = 4

_handler_count = 0
_timer_counter = 0

def update_timer_counter():
    global _timer_counter
    if event_list.head:
        _timer_counter = event_list.head.time_next - time.monotonic()

class Event:
    def __init__(self, handler, time_period, immediate=False):
        self.handler = handler
        self.time_next = time.monotonic()
        if not immediate:
            self.time_next += time_period
        self.time_period = time_period
        self.next = None

class EventList:
    def __init__(self):
        self.head = None

    def add(self, event):
        if not self.head or event.time_next < self.head.time_next:
            event.next = self.head
            self.head = event
            update_timer_counter()
        else:
            current = self.head
            while current.next:
                if current.next.time_next > event.time_next: break
                current = current.next
            event.next = current.next
            current.next = event

    def print_event_list(self):
        current = self.head
        while current:
            print(current)
            current = current.next

    def remove(self, handler):
        previous = None
        current = self.head
        while current:
            if current.handler == handler:
                if previous:
                    previous.next = current.next
                else:
                    self.head = current.next
                    update_timer_counter()
                break
            previous = current
            current = current.next
        return current

    def reset(self):
        current = self.head
        current_time = time.monotonic()
        while current:
            current.time_next = current_time + current.time_period
            current = current.next
        update_timer_counter()

    def update(self):
        if self.head:
            event = self.head
            event.time_next += event.time_period
            if event.next:
                if event.time_next > event.next.time_next:
                    self.head = event.next
                    self.add(event)
            update_timer_counter()


event_enabled = False
event_list = EventList()
event_loop_lock = Lock(f"{__name__}.loop")
event_loop_running = False
event_queue: "queue.Queue[Tuple[Any, Any]]" = queue.Queue()
flatout_handlers = []
mailboxes = OrderedDict()
queue_handlers = {}

def add_flatout_handler(handler):
    global _handler_count
    flatout_handlers.append(handler)
    _handler_count += 1

def remove_flatout_handler(handler):
    global _handler_count
    flatout_handlers.remove(handler)
    _handler_count -= 1

class Mailbox:
    def __init__(
        self, handler, name, increment_warning=_MAILBOX_INCREMENT_WARNING):

        self.handler = handler
        self.name = name
        self.increment_warning = increment_warning

        self.high_water_mark = 0
        self.last_warned_increment = 0
        self.queue = queue.Queue()
        self.size = 0

    def put(self, item):
        self.queue.put(item, block=False)
        self.size = self.queue.qsize()
        if self.size > self.high_water_mark:
            self.high_water_mark = self.size
        if self.size >= self.last_warned_increment + self.increment_warning:
            message = f"Mailbox {self.name}: size={self.size}"
            message = f"\033[91m{message}\033[0m"  # highlight red
        #   print(message)
            self.last_warned_increment += self.increment_warning

# First mailbox added has priority handling for all posted messages

def add_mailbox_handler(
    mailbox_handler,
    mailbox_name,
    mailbox_increment_warning=_MAILBOX_INCREMENT_WARNING):

    global _handler_count
    if mailbox_name not in mailboxes:
        mailbox = Mailbox(
            mailbox_handler, mailbox_name, mailbox_increment_warning)
        mailboxes[mailbox_name] = mailbox
        _handler_count += 1
    else:
        raise RuntimeError(f"Mailbox {mailbox_name}: Already exists")

def remove_mailbox_handler(mailbox_handler, mailbox_name):
    global _handler_count
    if mailbox_name in mailboxes:
        del(mailboxes[mailbox_name])
        _handler_count -= 1

def mailbox_put(mailbox_name, item):
    if mailbox_name in mailboxes:
        item = (item, time.monotonic())
        mailboxes[mailbox_name].put(item)
    else:
        raise RuntimeError(f"Mailbox {mailbox_name}: Not found")

def add_queue_handler(queue_handler, item_types=["default"]):
    global _handler_count
    for item_type in item_types:
        if not item_type in queue_handlers:
            queue_handlers[item_type] = []
        queue_handlers[item_type].append(queue_handler)
        _handler_count += 1

def remove_queue_handler(queue_handler, item_types=["default"]):
    global _handler_count
    for item_type in item_types:
        if item_type in queue_handlers:
            if queue_handler in queue_handlers[item_type]:
                queue_handlers[item_type].remove(queue_handler)
                _handler_count -= 1
            if len(queue_handlers[item_type]) == 0:
                del queue_handlers[item_type]

def queue_put(item, item_type="default"):
    event_queue.put((item, item_type))

def add_timer_handler(handler, time_period, immediate=False):
    global _handler_count
    event = Event(handler, time_period, immediate)
    event_list.add(event)
    _handler_count += 1

def remove_timer_handler(handler):
    global _handler_count
    event_list.remove(handler)
    _handler_count -= 1

def loop(loop_when_no_handlers=False):
    global event_enabled, event_loop_running, _timer_counter

    try:
        event_loop_lock.acquire("loop() #1")
        if event_loop_running:
            return
        event_loop_running = True
    finally:
        event_loop_lock.release()

    event_list.reset()

    try:
        event_enabled = True
        while event_enabled and (loop_when_no_handlers or _handler_count):
            event = event_list.head
            if event and _timer_counter <= 0:
                if time.monotonic() >= event.time_next:
                    event.handler()
                    event_list.update()
            sleep_time = 0.01

            if event_queue.qsize():
                (item, item_type) = event_queue.get()
                if item_type in queue_handlers:
                    for queue_handler in queue_handlers[item_type]:
                        queue_handler(item, item_type)

            if len(mailboxes) > 0:
                priority_mailbox = mailboxes[list(mailboxes)[0]]
                handle_mailboxes = True
                while handle_mailboxes:
                # TODO: Determine performance impact of "mailboxes.copy()"
                    for mailbox_name, mailbox in mailboxes.copy().items():
                        while mailbox.queue.qsize() > 0:
                            item, time_posted = mailbox.queue.get()
                            mailbox.handler(mailbox_name, item, time_posted)
                            if mailbox.queue != priority_mailbox.queue:
                                if priority_mailbox.queue.qsize() > 0:
                                    break
                        if priority_mailbox.queue.qsize() > 0:
                            break
                    handle_mailboxes = False

            if len(flatout_handlers):
                time_start = time.monotonic()
                for flatout_handler in flatout_handlers:
                    flatout_handler()
                sleep_time = sleep_time - (time.monotonic() - time_start)
# TODO:         _timer_counter -= sleep_time  # and don't sleep !
            if sleep_time > 0:
                time.sleep(sleep_time)
                _timer_counter -= sleep_time
    except KeyboardInterrupt:
        raise SystemExit("KeyboardInterrupt: abort !")
    finally:
        try:
            event_loop_lock.acquire("loop() #2")
            event_loop_running = False
        finally:
            event_loop_lock.release()

def terminate():
    global event_enabled
    event_enabled = False



#!/usr/bin/env python3
#
# Usage
# ~~~~~
# AIKO_LOG_LEVEL=DEBUG ./transport/test_mqtt.py create
#
# AIKO_LOG_LEVEL=DEBUG ./transport/test_mqtt.py test hello
#
# TODO: ./transport/test_mqtt.py delete
# TODO: ./transport/test_mqtt.py list [filter]
#
# To Do
# ~~~~~
# * Rename MQTTTest --> MQTTActorTest
#   Rename "test_mqtt.py" --> "mqtt_actor_test.py"
#
# - transport_mqtt.py: Complete create_actor_mqtt() and delete_actor_mqtt()
# - transport_mqtt.py: Wrap in CLI / TUI for handy tools for finding Services

from abc import abstractmethod
import click

from aiko_services.main import *

_VERSION = 0

ACTOR_TYPE = "mqtt_test"
PROTOCOL = f"{SERVICE_PROTOCOL_AIKO}/{ACTOR_TYPE}:{_VERSION}"

_LOGGER = aiko.logger(__name__)

# --------------------------------------------------------------------------- #

class MQTTTest(Actor):
    Interface.default("MQTTTest",
        "aiko_services.main.transport.test_mqtt.MQTTTestImpl")

    @abstractmethod
    def test(self, message):
        pass

class MQTTTestImpl(MQTTTest):
    def __init__(self, context):
        context.get_implementation("Actor").__init__(self, context)

        self.share["source_file"] = f"v{_VERSION}⇒ {__file__}"
        self.share["message"] = None

        self.actor_discovery = ActorDiscovery(self)
        tags = "*"  # ["class=AlohaHonuaActor"]  # TODO: CLI parameter
        filter = ServiceFilter("*", "*", "*", "*", "*", tags)
        self.actor_discovery.add_handler(self._actor_change_handler, filter)

    def _actor_change_handler(self, command, service_details):
        if command == "sync":
            _LOGGER.debug("sync")
        #   actors = self.actor_discovery.share_actor_mqtt(self.filter)
        #   name = "aiko/zeus/3704321.AlohaHonua"
        #   actor = self.actor_discovery.get_actor_mqtt(name)
        #   actor_topic = ".".join(name.split(".")[:-1])
        else:
            _LOGGER.debug(f"{command}: {service_details}")

    def test(self, message):
        _LOGGER.info(f"{self.name}: test({message})")
        self.ec_producer.update("message", message)
        payload_out = f"(test {message})"
        aiko.message.publish(self.topic_out, payload_out)

# --------------------------------------------------------------------------- #

@click.group()

def main():
    pass

@main.command(help="Transport MQTT Test Actor")

def create():
    init_args = actor_args(ACTOR_TYPE, protocol=PROTOCOL)
    mqtt_test = compose_instance(MQTTTestImpl, init_args)
    mqtt_test.run()

@main.command(name="test", help="Invoke function call on Transport MQTT Actor")
@click.argument("message", default=None, required=True)

def test(message):
    do_command(MQTTTest, ServiceFilter("*", "*", PROTOCOL, "*", "*", "*"),
        lambda mqtt_test: mqtt_test.test(message), terminate=True)
    aiko.process.run()

if __name__ == "__main__":
    main()

# --------------------------------------------------------------------------- #



from typing import Optional

from ...core import GQLBaseModel

__all__ = ["DatumSource"]


class DatumSource(GQLBaseModel):
    """How did a piece-of-data 'datum' come to be."""

    frame_id: Optional[int] = None
    host_id: Optional[str] = None
    pipeline_element_name: Optional[str] = None
    training_run_id: Optional[int] = None
    confidence: float

    def to_json(self):
        return self.model_dump()

    def serialize(self):
        return self.to_json()



from typing import Any, Optional, Type, Union
from uuid import UUID, uuid4

from pydantic import GetCoreSchemaHandler
from pydantic_core import core_schema

__all__ = ["LabeledUUID"]


class LabeledUUID(UUID):
    """LabeledUUID behaves exactly like a UUID, but has
    an additional 'label' field that is added to .__repr__() and .pretty()
    for readability
    """

    def __init__(self, *args, label: Optional[str] = None, **kwargs):
        if args and isinstance(args[0], UUID):
            args = list(args)
            args[0] = str(args[0])

        if args and isinstance(args[0], str):
            args = list(args)
            s = args[0]
            parts = s.split("|")
            if len(parts) == 2:
                uuid_str, label = parts
                args[0] = uuid_str
            elif len(parts) == 1:
                uuid_str = parts[0]
                args[0] = uuid_str
        super().__init__(*args, **kwargs)

        # Why not simply 'self.label = ... ?' I hear you ask.
        # The base class UUID is immutable so it blocks __setattr__, see:
        #   https://github.com/python/cpython/blob/main/Lib/uuid.py#L277
        # This is the way they do it in the UUID base class' constructor, see:
        #   https://github.com/python/cpython/blob/main/Lib/uuid.py#L222
        object.__setattr__(self, "label", "" if label is None else label)

    def __repr__(self) -> str:
        return self.short_str()

    def __eq__(self, other) -> bool:
        if isinstance(other, str):
            return str(self) == other
        else:
            return super().__eq__(other)

    def __hash__(self) -> int:
        # Is this the same as not overriding __hash__? You'd think that
        # wouldn't you, but no __hash__ is special. See:
        #
        #   https://stackoverflow.com/a/53519136
        return super().__hash__()

    @classmethod
    def uuid4(cls, label: Optional[str] = None):
        return cls(int=uuid4().int, label=label)

    @classmethod
    def from_str(cls, s):
        """Init from formatted string

        Args:
            s: 01234567-0123-0123-0123-012345678901|My label
        """
        uuid_str, label = s.split("|")
        return cls(uuid_str, label=label)

    @classmethod
    def from_dict(cls, *, uuid: Union[str, UUID], label: str):
        if isinstance(uuid, UUID):
            return cls(str(uuid), label=label)
        assert isinstance(uuid, str)
        return cls(uuid, label=label)

    def short_str(self) -> str:
        """1234..8901|My label"""
        uuid_str = str(self)
        short_uuid = f"{uuid_str[:4]}..{uuid_str[-4:]}"
        label = getattr(self, "label", "")
        return f"{short_uuid}|{label}" if label else short_uuid

    def verbose_str(self) -> str:
        """01234567-0123-0123-0123-012345678901|My label"""
        uuid_str = str(self)
        short_uuid = f"{uuid_str[:4]}..{uuid_str[-4:]}"
        label = getattr(self, "label", "")
        return f"{uuid_str}|{label}" if label else short_uuid

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: Type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        assert source is LabeledUUID

        # Union schema can handle
        # str, uuid or dict
        # if dict, the 'uuid' value can be a str or uuid object
        custom_schema = core_schema.union_schema(
            [
                core_schema.str_schema(),  # Handle string inputs
                core_schema.uuid_schema(),  # Handle UUID inputs
                core_schema.typed_dict_schema(
                    {
                        "uuid": core_schema.typed_dict_field(
                            core_schema.union_schema([core_schema.str_schema(), core_schema.uuid_schema()])
                        ),
                        "label": core_schema.typed_dict_field(core_schema.str_schema()),
                    }
                ),
            ]
        )
        return core_schema.no_info_after_validator_function(cls.validate, custom_schema)

    @classmethod
    def validate(cls, value: Any) -> "LabeledUUID":
        if isinstance(value, cls):
            return value
        elif isinstance(value, (str, UUID)):
            return LabeledUUID(value)
        elif isinstance(value, dict):
            assert "uuid" in value
            return LabeledUUID.from_dict(**value)
        else:
            raise ValueError(f"Invalid LabeledUUID {value}")

    @staticmethod
    def _serialize(value: "LabeledUUID") -> str:
        return str(value)



import re
from collections.abc import MutableMapping
from typing import Dict, List, Optional, Type, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict

__all__ = [
    "BelongsTo",
    "GQLBaseModel",
    "HLDataModel",
    "HLModelMap",
    "HasMany",
    "camel_to_snake",
    "snake_to_camel",
]


def camel_to_snake(s):
    if s.isupper():
        return s
    snake = re.sub("([A-Z])", r"_\g<1>", s).lower()
    if snake.startswith("_"):
        snake = snake[1:]
    return snake


def snake_to_camel(s):
    if s.isupper():
        return s
    first, *rest = s.split("_")
    return first + "".join(x.capitalize() for x in rest)


def map_keys_deeply(x, f):
    if isinstance(x, dict):
        result = {f(k): map_keys_deeply(v, f) for k, v in x.items()}
    elif isinstance(x, list):
        result = [map_keys_deeply(element, f) for element in x]
    else:
        result = x
    return result


class IdentityMap:
    def __init__(self):
        self._objects = {}

    def get(self, obj_cls, obj_id):
        return self._objects.get((obj_cls, obj_id))

    def add(self, obj):
        obj_cls = type(obj)
        obj_id = obj.get_id()
        self._objects[(obj_cls, obj_id)] = obj

    def remove(self, obj_cls, obj_id):
        _ = self._objects.pop((obj_cls, obj_id), None)


class GQLBaseModel(BaseModel):
    """To allow for seamless integration with the GraphQL CamelCaseWorld and
    th_python_snake_case_world. We have this customized Pydantic BaseModel.

    This expects all fields are defined in snake_case.
    When serialized the resulting dict will have CamelCase keys,
    this allows it to play nice with GraphQL.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    def __init__(self, *args, **kwargs):
        kwargs = map_keys_deeply(kwargs, camel_to_snake)
        super().__init__(*args, **kwargs)

    def gql_dict(self):
        d = self.model_dump(exclude_none=True)
        d = map_keys_deeply(d, snake_to_camel)
        return d


class HLDataModel(GQLBaseModel):
    """Adds the IdentityMap pattern to GQLBaseModel provide an API that ensures
    only one instance of an object is being accessed and modified in the code.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid", validate_assignment=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._get_identity_map().add(self)
        self._get_has_many_relationships()

    @classmethod
    def _get_identity_map(cls) -> IdentityMap:
        """Get the identity map for the current class (subclass)."""
        if not hasattr(cls, "_identity_map"):
            # Create a new identity map for the subclass
            cls._identity_map = IdentityMap()
        return cls._identity_map

    @classmethod
    def _add_has_many_relationship(cls, field, target_cls, back_populates):
        cls._get_has_many_relationships().append(
            {
                "field": field,
                "target_cls": target_cls,
                "back_populates": back_populates,
            }
        )

    @classmethod
    def _get_has_many_relationships(cls) -> List:
        if not hasattr(cls, "_has_many_relationships"):
            # Create a new identity map for the subclass
            cls._has_many_relationships = []
        return cls._has_many_relationships

    def get_id(self):
        """subclasses must have an .id member. Or you need to override this method"""
        try:
            return self.id  # type: ignore
        except AttributeError as _:
            raise AttributeError(
                f"{self.__class__.__name__} has no member self.id. "
                "Did you need to override def get_id(self)?"
            )

    @classmethod
    def find_by_id(cls, obj_id):
        return cls._get_identity_map().get(cls, obj_id)

    @classmethod
    def remove_by_id(cls, obj_id):
        return cls._get_identity_map().remove(cls, obj_id)


class HLModelMap(MutableMapping):

    def __init__(
        self,
        t: Union[Type, List],
        owner: Optional[HLDataModel] = None,
        back_populates: Optional[str] = None,
    ) -> None:
        """A typed dict that also does the book keeping to make sure
        references to the owner object is maintained

        Args
        """
        self._dict = {}

        self.owner = owner
        self.back_populates = back_populates

        # We may be dealing with repeatededly wrapped classes
        # so we use this to ensure we're adding properties to
        # the correct classes
        if isinstance(t, type):
            self._t = t
        elif _t := getattr(t, "original_class", None):
            # In this case 't' is an instance of the HasMany decorator
            self._t = _t
        elif isinstance(t, list):
            self._t = type(t[0])
            for item in t:
                self.add(item)
        else:
            raise ValueError("Could not infer HLModelMap type")

    def _validate_item(self, item) -> Union[UUID, int, str]:
        if not isinstance(item, self._t):
            raise TypeError(f"Invalid type, expected {self._t} got {type(item)}")

        id = item.get_id()
        return id

    def add(self, item):
        """Add an item to the HLModelMap and update the associated
        references
        """

        self._validate_item(item)

        if (self.owner is not None) and (self.back_populates is not None):
            # if isinstance(self.target_cls, str):
            #    target_cls = import_attr(self.target_cls)
            #    assert isinstance(target_cls, type)
            #    self.target_cls = target_cls

            # owner_inst = self.target_cls.find_by_id(self.owner_id)
            # if self.owner.get_id() != item.get_id():
            #    setattr(item, self.back_populates, self.owner)
            setattr(item, self.back_populates, self.owner)

        self._dict[item.get_id()] = item

    def remove(self, item):
        id = self._validate_item(item)
        del self._dict[id]

    def items(self):
        return self._dict.items()

    def keys(self):
        return list(self._dict.keys())

    def values(self):
        return list(self._dict.values())

    def clear(self):
        self._dict = {}

    def __len__(self):
        return len(self._dict)

    def __getitem__(self, id):
        return self._dict[id]

    def __setitem__(self, id, item):
        assert id == item.get_id()
        self.add(item)

    def __contains__(self, item):
        return (item.get_id() in self._dict) and (self._dict[item.get_id()] is item)

    def __repr__(self) -> str:
        return str(self._dict)

    def __delitem__(self, item):
        self.remove(item)

    def __iter__(self):
        return iter(self.values())


def HasMany(
    field,
    target_cls,
    back_populates,
):
    """

    Args:
        target_cls:   Import string to the class that 'is the many' class.
        field: The field of on the owner class that has the HLModelMap containing the target_cls instnaces
        back_populates: The name of the field on the target_cls that refers back to the owner instance

    Example:

        @HasMany("highlighter.client.Annotation", "annotations", "entity")
        class Entity(HLDataModel):
            id: UUID
            annotations: HLModelMap = ...

        class Annotation(HLDataModel):
            id: UUID
            entity: Entity  # Not 100% correct, but illustrative, see note.


        Note: We would normally use the 'BelongsTo' decorator on the Annotation class
        that would add and 'entity' property to the Annotation class that looks
        up the Entity instance by entity_id.
    """

    class Decorator:
        _is_decorator = True

        def __init__(self, cls):
            # We may be dealing with repeatededly wrapped classes
            # so we use this to ensure we're adding properties to
            # the correct classes
            self.cls = getattr(cls, "original_class", cls)
            self.cls._add_has_many_relationship(field, target_cls, back_populates)

        def __call__(self, *args, **kwargs):
            def transform_list_to_hlmodelmap(value):
                if isinstance(value, list):
                    return HLModelMap(value)
                return value

            for rel in self.cls._get_has_many_relationships():
                if rel["field"] in kwargs:
                    kwargs[rel["field"]] = transform_list_to_hlmodelmap(kwargs[rel["field"]])

            instance = self.cls(*args, **kwargs)
            for rel in self.cls._get_has_many_relationships():
                # TODO: instantiate field in the future
                field = getattr(instance, rel["field"])
                setattr(field, "owner", instance)
                setattr(field, "back_populates", rel["back_populates"])
            return instance

        def __getattr__(self, name):
            # Forward attribute access to the class (for class methods and attributes)
            return getattr(self.cls, name)

        @property
        def original_class(self):
            """Access the original, unwrapped class by following the __wrapped__ chain"""
            cls = self.cls
            # Follow the __wrapped__ chain until we hit the original class
            while hasattr(cls, "__wrapped__"):
                cls = cls.__wrapped__
            return cls

    return Decorator


def import_attr(item):
    import importlib

    parts = item.split(".")
    module_name = ".".join(parts[:-1])
    class_name = parts[-1]
    module = importlib.import_module(module_name)
    return getattr(module, class_name)


def BelongsTo(
    field: str,
    target_cls: Union[type, str],
    back_populates: str,
    back_populates_id_field: Optional[str] = None,
):

    if back_populates_id_field is None:
        back_populates_id_field = f"{field}_id"

    def decorator(
        cls,
        field=field,
        target_cls=target_cls,
        back_populates=back_populates,
        back_populates_id_field=back_populates_id_field,
    ):
        """

        Args:
            field: The name of the property to add to the decorated class the
            will lookup an instnace to the owner.

            target_cls: Import path to the cls that the decorated class belongs to

            back_populates: The fiend name on the owner that refers to a
            collection of the decorated class. Must be a HLModelMap.

            back_populates_id_field: The field name on the decorated class
            that refers to the owner. If None will be f'{field}_id'


        @BelongsTo(
                   "entity",
                   "highlighter.client.Entity",
                   "annotations",
                   )
        class Annotation():
          id: UUID
          entity_id: UUID

        class Entity():
          id: UUID
          annotations: HLModelMap(Annotation) ...

        """
        _cls_name = target_cls.split(".")[-1]
        if field is None:
            field = camel_to_snake(_cls_name)

        def _get_belongs_to_class(
            target_cls=target_cls,
            back_populates=back_populates,
            field=field,
            back_populates_id_field=back_populates_id_field,
        ):
            _cls = import_attr(target_cls)
            if back_populates_id_field is None:
                back_populates_id_field = f"{field}_id"
            return _cls, back_populates, field, back_populates_id_field

        # Equivalent to:
        # @property
        # def entity(self):
        #    if self.entity_id is None:
        #        return None
        #    return Entity.find_by_id(self.entity_id)
        def _getter(self):
            target_cls, _, _, back_populates_id_field = _get_belongs_to_class()
            id = getattr(self, back_populates_id_field)
            if id is None:
                return None
            return target_cls.find_by_id(id)

        # @entity.setter
        # def entity(self, v: Optional["Entity"]):
        #    if self.entity_id is None:
        #        if v is not None:
        #            self.entity_id = v.id
        #            Entity.find_by_id(v.id).annotations.add(self)

        #    else:
        #        if v is None:
        #            if self in Entity.find_by_id(self.entity_id).annotations:
        #                Entity.find_by_id(self.entity_id).annotations.remove(self)
        #            self.entity_id = None

        #        elif self not in Entity.find_by_id(v.id).annotations:
        #            self.entity_id = v.id
        #            Entity.find_by_id(v.id).annotations.add(self)
        def _setter(self, v):
            if getattr(self, "_add_in_progress", False):
                return
            target_cls, _, _, back_populates_id_field = _get_belongs_to_class()
            id = getattr(self, back_populates_id_field)
            if id is None:
                if v is not None:
                    setattr(self, back_populates_id_field, v.get_id())
                    new_belongs_to_inst = getattr(target_cls, "find_by_id")(v.get_id())

                    self._add_in_progress = True
                    getattr(new_belongs_to_inst, back_populates).add(self)
                    self._add_in_progress = False

            else:
                if v is None:
                    old_belongs_to_inst = getattr(target_cls, "find_by_id")(id)
                    old_belongs_to_collection = getattr(old_belongs_to_inst, back_populates)
                    if self in old_belongs_to_collection:
                        old_belongs_to_collection.remove(self)
                    setattr(self, back_populates_id_field, None)

                else:
                    new_belongs_to_inst = getattr(target_cls, "find_by_id")(v.get_id())
                    new_belongs_to_collection = getattr(new_belongs_to_inst, back_populates)
                    if self not in new_belongs_to_collection:
                        setattr(self, back_populates_id_field, v.get_id())

                        self._add_in_progress = True
                        new_belongs_to_collection.add(self)
                        self._add_in_progress = False

                    old_belongs_to_inst = getattr(target_cls, "find_by_id")(id)
                    old_belongs_to_collection = getattr(old_belongs_to_inst, back_populates)
                    if self in old_belongs_to_collection:
                        old_belongs_to_collection.remove(self)

        # We may be dealing with repeatededly wrapped classes
        # so we use this to ensure we're adding properties to
        # the correct classes
        original_class = getattr(cls, "original_class", cls)
        setattr(original_class, field, property(_getter, _setter))

        return original_class

    return decorator



from typing import Any, Dict, List, Tuple
from uuid import UUID

from highlighter import DataFile, DatumSource, Entity, Observation
from highlighter.agent.capabilities import Capability, StreamEvent


class MockClassifier(Capability):
    """Mock classifier that creates a dummy observation"""

    def process_frame(
        self, stream, entities: Dict[UUID, Entity], data_files: List[DataFile]
    ) -> Tuple[StreamEvent, dict]:
        # When we specify the output taxonomy of the element in the
        # pipeline element outputs then we can get it from there. For now
        # we add it to the parameters
        output_attribute_id, found = self.get_parameter("output_attribute_id")
        if not found:
            raise ValueError("'output_attribute_id' not in element parameters")
        self.output_attribute_id = UUID(output_attribute_id)
        output_enum_id, found = self.get_parameter("output_enum_id")
        if not found:
            raise ValueError("'output_enum_id' not in element parameters")
        self.output_enum_id = UUID(output_enum_id)

        for entity_id, entity in entities.items():
            for annotation in entity.annotations:
                datum_source = DatumSource(confidence=1, frame_id=annotation.datum_source.frame_id)
                obs = Observation(
                    attribute_id=self.output_attribute_id,
                    value=self.output_enum_id,
                    datum_source=datum_source,
                )
                obs.entity_id = entity_id
                annotation.observations.add(obs)
        return StreamEvent.OKAY, {"entities": entities}



# Usage
# ~~~~~
# pytest [-s] unit/test_hook.py
# pytest [-s] unit/test_hook.py::test_hook
#
# test_hook() performs ...
# - Add Hook used within the Framework "component" code
# - Add HookHandler created by the third-party developer
# - Run Hook --> HookHandlers
# - Remove HookHandler
# - Remove Hook
#
# Where "# by framework" is code inserted into the Aiko Services framework
# Where "# by developer" is code that utilizes that framework hook
#
# To Do
# ~~~~~
# - None, yet !

import aiko_services as aiko

HOOK_NAME = "HookTest.test_hook:0"

class HookTest(aiko.Actor):  # primarily acts as a Framework component
    def __init__(self, context):
        context.get_implementation("Actor").__init__(self, context)

    def hook_check(self):
        hook_count_base = len(self.get_hooks())
        self.variable_0 = 1
        self.variable_1 = "test"

        self.add_hook(HOOK_NAME)                                 # by framework
        hooks = self.get_hooks()
        assert len(hooks) == hook_count_base + 1

        self.add_hook_handler(HOOK_NAME, self.hook_handler)      # by developer
        hook_test = self.get_hook(HOOK_NAME)
        hook_test_handlers = hook_test.handlers
        assert len(hook_test_handlers) == 1

        self.run_hook(HOOK_NAME, lambda: {                       # by framework
            "variable_0": self.variable_0,
            "variable_1": self.variable_1})

        self.remove_hook_handler(HOOK_NAME, self.hook_handler)   # by developer
        assert len(hook_test_handlers) == 0

        self.remove_hook(HOOK_NAME)                              # by framework
        assert len(hooks) == hook_count_base

        aiko.process.terminate()  # TODO: Improve Aiko Services Process exit

    def hook_handler(self, name, component, logger, variables, options):
        assert name == HOOK_NAME                                 # by developer
        assert component == self
        assert logger is not None
        assert variables["variable_0"] == component.variable_0
        assert variables["variable_1"] == component.variable_1
        assert options == {}

def test_hook():                                                 # by developer
    init_args = aiko.actor_args("test_hook")
    hook_test = aiko.compose_instance(HookTest, init_args)
    aiko.event.add_timer_handler(hook_test.hook_check, 0.0)
    aiko.process.run(mqtt_connection_required=False)



from datetime import datetime, timezone
from typing import Any, List, Optional, Union
from uuid import UUID, uuid4

import numpy as np
import shapely.geometry as geom
from pydantic import ConfigDict, Field

from highlighter.client.base_models.base_models import Polygon
from highlighter.core.hl_base_model import BelongsTo

from ...core import (
    DATA_FILE_ATTRIBUTE_UUID,
    EMBEDDING_ATTRIBUTE_UUID,
    OBJECT_CLASS_ATTRIBUTE_UUID,
    PIXEL_LOCATION_ATTRIBUTE_UUID,
    TRACK_ATTRIBUTE_UUID,
    HLDataModel,
    LabeledUUID,
)
from .datum_source import DatumSource

__all__ = ["Observation"]


@BelongsTo(
    "entity",
    target_cls="highlighter.client.Entity",
    back_populates="global_observations",
)
@BelongsTo(
    "annotation",
    target_cls="highlighter.client.Annotation",
    back_populates="observations",
)
class Observation(HLDataModel):
    """
    entity_id and attribute_id are global
    value is tied to the attribute, and we have unit for it, so it doesn't appear here
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    id: UUID = Field(..., default_factory=uuid4)
    entity_id: Optional[UUID] = None
    annotation_id: Optional[UUID] = None
    attribute_id: LabeledUUID
    value: Any  # <-- ToDo: Add specific types
    occurred_at: datetime = Field(..., default_factory=lambda: datetime.now(timezone.utc))
    datum_source: DatumSource
    unit: Optional[str] = None
    file_id: Optional[UUID] = None

    @classmethod
    def from_deprecated_eavt(cls, eavt, id: UUID, annotation_id: Optional[UUID] = None):
        return cls(
            id=id,
            entity_id=eavt.entity_id,
            annotation_id=annotation_id,
            attribute_id=eavt.attribute_id,
            value=eavt.value,
            occurred_at=eavt.time,
            datum_source=eavt.datum_source,
            unit=eavt.unit,
            file_id=eavt.file_id,
        )

    def to_json(self):
        data = self.model_dump()
        data["id"] = str(self.id)
        data["value"] = self.value.to_json() if hasattr(self.value, "to_json") else self.value
        data["file_id"] = str(self.file_id)
        return data

    def model_dump(self, *args, **kwargs):
        value = self.value
        if isinstance(value, Polygon):
            value = value.dict()

        if isinstance(value, UUID):
            value = str(value)

        return dict(
            entity_id=str(self.entity_id),
            attribute_id=str(self.attribute_id),
            value=value,
            occurred_at=self.occurred_at.isoformat(),
            datum_source=self.datum_source.model_dump(),
        )

    def gql_dict(self):
        d = super().gql_dict()
        d["time"] = d.pop("occurredAt")
        return d

    def serialize(self):
        return self.to_json()

    def is_pixel_location(self):
        return str(self.attribute_id) == PIXEL_LOCATION_ATTRIBUTE_UUID

    def is_object_class(self):
        return str(self.attribute_id) == OBJECT_CLASS_ATTRIBUTE_UUID

    def is_track(self):
        return str(self.attribute_id) == TRACK_ATTRIBUTE_UUID

    def is_embedding(self):
        return str(self.attribute_id) == EMBEDDING_ATTRIBUTE_UUID

    def get_confidence(self):
        return self.datum_source.confidence

    @classmethod
    def make_scalar_observation(
        cls,
        value: Union[int, float, tuple, list],
        attribute_id: LabeledUUID,
        occurred_at: datetime,
        pipeline_element_name: Optional[str] = None,
        training_run_id: Optional[int] = None,
        host_id: Optional[str] = None,
        frame_id: Optional[int] = None,
        unit: Optional[str] = None,
    ):
        datum_source = DatumSource(
            confidence=1.0,
            pipeline_element_name=pipeline_element_name,
            training_run_id=training_run_id,
            host_id=host_id,
            frame_id=frame_id,
        )
        if isinstance(value, tuple):
            value = list(value)

        return cls(
            attribute_id=attribute_id,
            value=value,
            datum_source=datum_source,
            occurred_at=occurred_at,
            unit=unit,
        )

    @classmethod
    def make_image_observation(
        cls,
        image: np.ndarray,
        occurred_at: datetime,
        pipeline_element_name: Optional[str] = None,
        training_run_id: Optional[int] = None,
        host_id: Optional[str] = None,
        frame_id: Optional[int] = None,
    ):
        datum_source = DatumSource(
            confidence=1.0,
            pipeline_element_name=pipeline_element_name,
            training_run_id=training_run_id,
            host_id=host_id,
            frame_id=frame_id,
        )
        return cls(
            attribute_id=DATA_FILE_ATTRIBUTE_UUID,
            value=image,
            datum_source=datum_source,
            occurred_at=occurred_at,
        )

    @classmethod
    def make_embedding_observation(
        cls,
        embedding: List[float],
        occurred_at: datetime,
        pipeline_element_name: Optional[str] = None,
        training_run_id: Optional[int] = None,
        host_id: Optional[str] = None,
        frame_id: Optional[int] = None,
    ):
        if not isinstance(embedding, list):
            t = type(embedding)
            raise ValueError(f"embedding must be list of float not {t}")

        datum_source = DatumSource(
            confidence=1.0,
            pipeline_element_name=pipeline_element_name,
            training_run_id=training_run_id,
            host_id=host_id,
            frame_id=frame_id,
        )
        return cls(
            attribute_id=EMBEDDING_ATTRIBUTE_UUID,
            value=embedding,
            datum_source=datum_source,
            occurred_at=occurred_at,
        )

    @classmethod
    def make_pixel_location_observation(
        cls,
        value: Union[
            geom.Polygon,
            geom.MultiPolygon,
            geom.LineString,
            geom.Point,
        ],
        confidence: float,
        occurred_at: datetime,
        pipeline_element_name: Optional[str] = None,
        training_run_id: Optional[int] = None,
        host_id: Optional[str] = None,
        frame_id: Optional[int] = None,
    ):
        """Create a new pixel_location attribute"""

        datum_source = DatumSource(
            confidence=confidence,
            pipeline_element_name=pipeline_element_name,
            training_run_id=training_run_id,
            host_id=host_id,
            frame_id=frame_id,
        )

        return cls(
            attribute_id=PIXEL_LOCATION_ATTRIBUTE_UUID,
            value=value,
            datum_source=datum_source,
            occurred_at=occurred_at,
        )

    @classmethod
    def make_enum_observation(
        cls,
        attribute_uuid: UUID,
        attribute_label: str,
        enum_value: str,
        enum_id: UUID,
        confidence: float,
        occurred_at: datetime,
        pipeline_element_name: Optional[str] = None,
        training_run_id: Optional[int] = None,
        host_id: Optional[str] = None,
        frame_id: Optional[int] = None,
    ):
        """Make an Observation with an enum attribute"""
        datum_source = DatumSource(
            confidence=confidence,
            pipeline_element_name=pipeline_element_name,
            training_run_id=training_run_id,
            host_id=host_id,
            frame_id=frame_id,
        )

        return cls(
            attribute_id=LabeledUUID(
                attribute_uuid,
                label=attribute_label,
            ),
            value=LabeledUUID(
                enum_id,
                label=enum_value,
            ),
            datum_source=datum_source,
            occurred_at=occurred_at,
        )

    @classmethod
    def make_object_class_observation(
        cls,
        object_class_uuid: UUID,
        object_class_value: str,
        confidence: float,
        occurred_at: datetime,
        pipeline_element_name: Optional[str] = None,
        training_run_id: Optional[int] = None,
        host_id: Optional[str] = None,
        frame_id: Optional[int] = None,
    ):
        """Convienence method to make an Observation with an object_class"""
        datum_source = DatumSource(
            confidence=confidence,
            pipeline_element_name=pipeline_element_name,
            training_run_id=training_run_id,
            host_id=host_id,
            frame_id=frame_id,
        )

        return cls(
            attribute_id=OBJECT_CLASS_ATTRIBUTE_UUID,
            value=LabeledUUID(
                object_class_uuid,
                label=object_class_value,
            ),
            datum_source=datum_source,
            occurred_at=occurred_at,
        )

    @classmethod
    def make_boolean_observation(
        cls,
        attribute_uuid: UUID,
        attribute_label: str,
        value: bool,
        confidence: float,
        occurred_at: datetime,
        pipeline_element_name: Optional[str] = None,
        training_run_id: Optional[int] = None,
        host_id: Optional[str] = None,
        frame_id: Optional[int] = None,
    ):
        """Convienence method to make an Observation with an object_class"""
        if not isinstance(value, bool):
            raise ValueError(
                "make_boolean_observation expects value arg to be of type bool "
                f"got: {value} of type: {type(value)}"
            )

        datum_source = DatumSource(
            confidence=confidence,
            pipeline_element_name=pipeline_element_name,
            training_run_id=training_run_id,
            host_id=host_id,
            frame_id=frame_id,
        )

        return cls(
            attribute_id=LabeledUUID(
                attribute_uuid,
                label=attribute_label,
            ),
            value=value,
            datum_source=datum_source,
            occurred_at=occurred_at,
        )

    @classmethod
    def make_detection_observation_pair(
        cls,
        location_value: Union[geom.Polygon, geom.MultiPolygon],
        object_class_value: str,
        object_class_uuid: UUID,
        confidence: float,
        occurred_at: datetime,
        pipeline_element_name: Optional[str] = None,
        training_run_id: Optional[int] = None,
        host_id: Optional[str] = None,
        frame_id: Optional[int] = None,
    ):
        """Convienence method to make both a pixel_location and
        object_class attribute, returning them both in a list
        """
        pixel_location_observation = Observation.make_pixel_location_observation(
            location_value,
            confidence,
            pipeline_element_name=pipeline_element_name,
            training_run_id=training_run_id,
            host_id=host_id,
            frame_id=frame_id,
            occurred_at=occurred_at,
        )

        object_class_observation = Observation.make_object_class_observation(
            object_class_uuid,
            object_class_value,
            confidence,
            pipeline_element_name=pipeline_element_name,
            training_run_id=training_run_id,
            host_id=host_id,
            frame_id=frame_id,
            occurred_at=occurred_at,
        )
        return [pixel_location_observation, object_class_observation]



from datetime import datetime
from typing import List, Tuple
from uuid import UUID, uuid4

from shapely import geometry as geom

from highlighter import (
    OBJECT_CLASS_ATTRIBUTE_UUID,
    Annotation,
    DataFile,
    DatumSource,
    Entity,
    Observation,
)
from highlighter.agent.capabilities import Capability, StreamEvent


class MockDetector(Capability):

    def process_frame(self, stream, data_files: List[DataFile]) -> Tuple[StreamEvent, dict]:
        """Mock detector that creates dummy detections on image or text source data"""

        output_object_class_ids, found = self.get_parameter("output_object_class_ids")
        if not found:
            raise ValueError("'output_object_class_ids' not in element parameters")
        self.output_object_class_ids = [UUID(id) for id in output_object_class_ids]

        if len(data_files) == 0:
            return StreamEvent.OKAY, {"entities": {}}

        entities = {}
        for file in data_files:
            if file.content_type.startswith("image"):
                locations = [
                    geom.Polygon([(0, 0), (0, 10), (10, 10), (10, 0)]),
                    geom.Polygon([(100, 100), (100, 110), (110, 110), (110, 100)]),
                ]
            elif file.content_type.startswith("text"):
                locations = [geom.LineString([(0, 0), (10, 0)]), geom.LineString([(100, 100), (110, 100)])]
            else:
                raise ValueError(
                    "MockDetector.process_frame() must be given either 'image' or 'text' data files"
                )
            for location in locations:
                for object_class_id in self.output_object_class_ids:
                    entity_id = uuid4()
                    entities[entity_id] = Entity(id=entity_id)

                    annotation_datum_source = DatumSource(confidence=1, frame_id=file.media_frame_index)
                    annotation = Annotation(
                        id=uuid4(),
                        location=location,
                        data_file_id=file.file_id,
                        datum_source=annotation_datum_source,
                    )
                    entities[entity_id].annotations.add(annotation)

                    observation_datum_source = DatumSource(confidence=1, frame_id=file.media_frame_index)
                    observation = Observation(
                        id=uuid4(),
                        attribute_id=OBJECT_CLASS_ATTRIBUTE_UUID,
                        value=object_class_id,
                        occurred_at=datetime.now(),
                        datum_source=observation_datum_source,
                    )
                    annotation.observations.add(observation)
        return StreamEvent.OKAY, {"entities": entities}



# Usage
# ~~~~~
# cd pavilion_farms_monitor  # repository top-level
# aiko_pipeline create agents/assessment_pipeline.json -s 1  # -ll debug
#
# To Do
# ~~~~~
# - Configure nix environment to include HL SDK
#   - Update this source code to use HL SDK instead of the mock methods below
#
# - Fix random Pipeline hangs when using ImageGather StreamEvent.DROP_FRAME :(
#   - Currently, using a solid work-around instead of StreamEvent.DROP_FRAME
#
# - Implement ImageToVideo parameter to decide whether the resulting video
#   is provided in-memory or as a file in the file-system
#
# - Decide whether to keep the ImageToVideo OpenCV implementation as an option ?

import io
import os
import threading
from datetime import datetime
from typing import Optional, Tuple

import aiko_services as aiko
import av

from highlighter.agent.capabilities import Capability
from highlighter.agent.capabilities.base_capability import StreamEvent
from highlighter.client import HLClient, assessments, data_files

__all__ = ["ImageGather"]

_GATHER_SIZE = 25 * 10  # frames * seconds


class ImageGather(Capability):
    class DefaultStreamParameters(Capability.DefaultStreamParameters):
        task_id: str
        data_source_id: int

        list_size: int = _GATHER_SIZE
        frame_rate: int = 25
        bit_rate: int = 1000000
        template: str = "video_{:05d}.mp4"
        resolution: Optional[str] = None

    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("image_gather:0")
        context.get_implementation("PipelineElement").__init__(self, context)
        self._threads = []
        client = HLClient.get_client()
        self._client_endpoint = client.endpoint_url
        self._client_token = client.api_token

    def _get_new_client(self) -> HLClient:
        return HLClient.from_credential(self._client_token, self._client_endpoint)

    def start_stream(self, stream, stream_id):
        self.start_time = datetime.now()
        self.chunk_start_time = self.start_time
        stream.variables["image_gather_list"] = []
        return aiko.StreamEvent.OKAY, {}

    def process_frame(self, stream, data_files) -> Tuple[StreamEvent, dict]:
        images = [(d.content, d.recorded_at) for d in data_files]
        list_size, _ = self.get_parameter("list_size")
        image_gather_list = stream.variables["image_gather_list"]
        if image_gather_list == []:
            prev = self.chunk_start_time
            self.chunk_start_time = images[0][1]
            print("first frame", self.chunk_start_time, self.chunk_start_time - prev)
        image_gather_list += images

        if len(image_gather_list) < list_size:
            return aiko.StreamEvent.OKAY, {"images": [], "start_time": None}  # TODO: Workaround :(

        stream.variables["image_gather_list"] = []
        self.logger.debug(f"{self.my_id()}: images gathered")
        self._start_upload(stream, image_gather_list)

        return aiko.StreamEvent.OKAY, {}

    def stop_stream(self, stream, stream_id):
        image_gather_list = stream.variables["image_gather_list"]

        if len(image_gather_list):
            self._start_upload(stream, image_gather_list)

        self.logger.debug(f"{self.my_id()}: waiting for {len(self._threads)} upload threads to complete")
        for t in self._threads:
            t.join()

        self._assessment_finalise(stream.variables)

        return StreamEvent.OKAY, {}

    def _start_upload(self, stream, image_gather_list):
        template, _ = self.get_parameter("filename_template")
        path = template.format(int(image_gather_list[0][1].timestamp()))
        thread = threading.Thread(
            target=self._images_to_video, args=(image_gather_list, path, stream.variables)
        )
        thread.start()
        self._threads.append(thread)

    def _images_to_video(self, images, path, cache):
        bit_rate, _ = self.get_parameter("bit_rate")
        frame_rate, _ = self.get_parameter("frame_rate")
        first_image = images[0][0]
        resolution_default = f"{first_image.shape[1]}x{first_image.shape[0]}"
        resolution, _ = self.get_parameter("resolution")
        if resolution is None:
            resolution = resolution_default
        width, height = resolution.split("x")

        buffer = io.BytesIO()
        container = av.open(buffer, mode="w", format="mp4")
        video_stream = container.add_stream("h264", rate=frame_rate)
        video_stream.bit_rate = bit_rate if bit_rate else video_stream.bit_rate
        video_stream.pix_fmt = "yuv420p"
        video_stream.width, video_stream.height = int(width), int(height)

        for timestamp, image in enumerate(images):
            #   video_frame = av.VideoFrame.from_image(pil_image)  # PIL Image
            video_frame = av.VideoFrame.from_ndarray(image[0])  # Numpy Image
            video_frame.pts = timestamp
            packet = video_stream.encode(video_frame)
            if packet is not None:
                container.mux(packet)

        final_packet = video_stream.encode(None)
        if final_packet is not None:
            container.mux(final_packet)
        container.close()

        video_mp4 = buffer.getvalue()
        try:
            with open(path, "wb") as file:
                file.write(video_mp4)
            self._assessment_append_paths(path, cache, images[0][1])
        finally:
            if os.path.exists(path):
                os.remove(path)
            else:
                self.logger.warning(f'{self.my_id()}: Assessment file: "{path}" not found')
        return path

    def _assessment_append(self, path, cache, start_time):
        data_source_id, found = self.get_parameter("data_source_id")
        if not found:
            raise KeyError('Must provide "data_source_id" parameter')

        task_id, found = self.get_parameter("task_id")
        if not found:
            raise KeyError('Must provide "task_id" parameter')

        hl_client = self._get_new_client()
        if "assessment" not in cache:

            data_file = data_files.create_data_file(
                hl_client, path, data_source_id, recorded_at=start_time.isoformat()
            )

            assessment = assessments.create_assessment_not_finalised(
                hl_client, task_id=str(task_id), image_id=data_file.id
            )
            cache["assessment"] = assessment
            assessments.append_data_files_to_not_finalised_assessment(
                hl_client, cache["assessment"], [data_file]
            )

        else:

            data_file = data_files.create_data_file(
                hl_client, path, data_source_id, recorded_at=start_time.isoformat()
            )

            assessments.append_data_files_to_not_finalised_assessment(
                hl_client, cache["assessment"], [data_file]
            )

    def _assessment_append_paths(self, path, cache, start_time):
        self.logger.info(f"{path} {start_time}")
        self._assessment_append(path, cache, start_time)

    def _assessment_finalise(self, cache):
        if "assessment" in cache:
            hl_client = self._get_new_client()
            assessments.finalise(hl_client, cache["assessment"])
            del cache["assessment"]



import mimetypes
from pathlib import Path
from typing import List, Tuple
from warnings import warn

from shapely.wkt import loads as wkt_loads

from ....client import PixelLocationAttributeValue
from ....core import OBJECT_CLASS_ATTRIBUTE_UUID
from ...base_models import AttributeRecord, ImageRecord
from ...interfaces import IReader

__all__ = [
    "HighlighterAssessmentsReader",
]


class HighlighterAssessmentsReader(IReader):
    format_name = "highlighter_assessments"

    def __init__(self, assessments_gen):
        self.assessments_gen = assessments_gen

    def read(self) -> Tuple[List[AttributeRecord], List[ImageRecord]]:
        attribute_records = []
        data_file_records = []
        for assessment in self.assessments_gen:
            data_file = assessment.image
            assessment_id = assessment.id
            hash_signature = assessment.hash_signature
            data_file_id = data_file.id
            filename_original = Path(data_file.original_source_url)

            ext = filename_original.suffix
            if ext == "":
                ext = mimetypes.guess_extension(data_file.mime_type)
                assert isinstance(ext, str)
                ext = ext.lower()

            filename = f"{data_file_id}{ext}"
            data_file_records.append(
                ImageRecord(
                    data_file_id=data_file_id,
                    width=data_file.width,
                    height=data_file.height,
                    filename=filename,
                    extra_fields={"filename_original": str(filename_original)},
                    assessment_id=assessment_id,
                    hash_signature=hash_signature,
                )
            )

            for eavt in assessment.entity_attribute_values:
                value = eavt.value
                if value is None:
                    value = eavt.entity_attribute_enum.id

                datum_source = eavt.entity_datum_source
                if datum_source is None:
                    conf = 1.0
                else:
                    conf = datum_source.confidence

                attribute_records.append(
                    AttributeRecord(
                        data_file_id=data_file_id,
                        entity_id=eavt.entity_id,
                        attribute_id=eavt.entity_attribute.id,
                        attribute_name=eavt.entity_attribute.name,
                        value=value,
                        confidence=conf,
                    )
                )

            for annotation in assessment.annotations:
                if annotation.location is None:
                    warn("Null value found in location. Get it together bro.")
                    continue

                confidence = getattr(annotation, "confidence", None)
                object_class = annotation.object_class
                attribute_records.append(
                    AttributeRecord(
                        data_file_id=data_file_id,
                        entity_id=annotation.entity_id,
                        attribute_id=str(OBJECT_CLASS_ATTRIBUTE_UUID),
                        attribute_name=OBJECT_CLASS_ATTRIBUTE_UUID.label,
                        value=object_class.uuid,
                        frame_id=annotation.frame_id,
                        confidence=confidence,
                    )
                )

                geometry = wkt_loads(annotation.location)
                pixel_location_attribute_value = PixelLocationAttributeValue.from_geom(
                    geometry, confidence=confidence
                )
                attribute_records.append(
                    AttributeRecord.from_attribute_value(
                        data_file_id,
                        pixel_location_attribute_value,
                        entity_id=annotation.entity_id,
                        frame_id=annotation.frame_id,
                    )
                )

        return data_file_records, attribute_records



import json

import click

from highlighter.core.logging import LOG

from ..client import EvaluationMetric, EvaluationMetricResult, HLJSONEncoder
from ..client.evaluation import (
    EvaluationMetricCodeEnum,
    create_evaluation_metric_result,
    find_or_create_evaluation_metric,
    get_existing_evaluation_metrics,
)


@click.group("evaluation")
@click.pass_context
def evaluation_group(ctx):
    pass


@evaluation_group.command("create-metric")
@click.argument("evaluation_id", type=int)
@click.argument(
    "code",
    type=click.Choice(tuple(EvaluationMetricCodeEnum.__members__.values())),
)
@click.argument("metric_name", type=str)
@click.option("--description", "-d", type=str, default=None)
@click.option(
    "--iou",
    type=click.FloatRange(min=0.0, max=1.0),
    default=None,
    help="Intersection-Over-Union. A measure of overlap",
)
@click.option("--weighted", type=bool, default=False, help="Is the metric weighted")
@click.option(
    "--object-class-uuid", "-o", type=str, default=None, help="The object class this metric refers to"
)
@click.pass_context
def create_metric(ctx, evaluation_id, code, metric_name, description, iou, weighted, object_class_uuid):
    """Create an evaluation metric record

    \b
    Args:
        evaluation_id: https://...highlighter.ai/research_plans/<EVALUATION_ID>
        metric_name: Human readable name
        code: Select from common metrics or use 'Other'

    \b
    Example:
        hl evaluation create-metric -d 'useful description' 123 Other foo
    """

    evaluation_metric: EvaluationMetric
    evaluation_metric, _ = find_or_create_evaluation_metric(
        ctx.obj["client"],
        evaluation_id,
        code,
        metric_name,
        description=description,
        iou=iou,
        weighted=weighted,
        object_class_uuid=object_class_uuid,
    )
    click.echo(json.dumps(evaluation_metric.model_dump(), cls=HLJSONEncoder))


@evaluation_group.command("create-result")
@click.argument("evaluation_metric_id", type=int)
@click.argument("result", type=float)
@click.option("--occured-at", type=click.DateTime(), default=None)
@click.option("--object-class-uuid", "-o", type=str, default=None)
@click.option("--training-run-id", "-t", type=int, default=None)
@click.option(
    "--lookup",
    "-l",
    type=click.Tuple((int, str)),
    default=None,
    help="Lookup the evaluation_metric_id, EVALUATION_ID, METRIC_NAME. Must set evaluation_metric_id to 0",
)
@click.pass_context
def create_result(ctx, evaluation_metric_id, result, occured_at, object_class_uuid, training_run_id, lookup):
    """Create an evaluation metric result

    \b
    Args:
        evaluation_metric_id: The id of the evaluation metric. To lookup by name
        instead, set this to 0 and use --lookup option
        result: float

    \b
    Examples:
        # Set an ExperimentResult with a known experiment_metric_id
        hl evaluation create-result 123 4.2
        \b
        # Set an ExperimentResult using the --lookup flag
        # we know the evaluation_id is 42 and the metric_name is 'Foo'
        # Note:
        #  - The metric_name is case sensitive
        #  - We set evaluation_metric_id to 0
        hl evaluation create-result --lookup 42 Foo 0 4.2


    """
    assert evaluation_metric_id >= 0, f"evaluation_metric_id must be >= 0, got: {evaluation_metric_id}"

    client = ctx.obj["client"]
    if evaluation_metric_id == 0:
        assert lookup is not None, "To lookup an EvaluationMetric by name you must set the " "--lookup option"

        evaluation_id, metric_name = lookup
        existing_evaluation_metrics = {
            r.name: r for r in get_existing_evaluation_metrics(client, evaluation_id)
        }
        evaluation_metric = existing_evaluation_metrics.get(metric_name, None)
        assert evaluation_metric is not None, (
            f"Could not find an EvaluationMetric named '{metric_name}' " f"in Evaluation({evaluation_id})"
        )
        LOG.info(f"Found existing metric: {evaluation_metric}")
        evaluation_metric_id = evaluation_metric.id

    assert isinstance(
        evaluation_metric_id, int
    ), f"evaluation_metric_id must be an int, got: {evaluation_metric_id}"

    metric_result: EvaluationMetricResult
    metric_result = create_evaluation_metric_result(
        client,
        evaluation_metric_id,
        result,
        occured_at=occured_at,
        object_class_uuid=object_class_uuid,
        training_run_id=training_run_id,
    )
    click.echo(json.dumps(metric_result.model_dump(), cls=HLJSONEncoder))



from functools import partial

from colorama import Back, Fore, Style


def message_in_color(color: str, message: str):
    return color + message + Fore.RESET


red = partial(message_in_color, Fore.RED)
green = partial(message_in_color, Fore.GREEN)


def message_in_color_front_and_back(front_color: str, back_color: str, style: str, message: str):
    message = front_color + message + Fore.RESET
    if back_color is None:
        return message

    message = back_color + message + Back.RESET
    if style is None:
        return message

    return style + message + Style.RESET_ALL


class ColoredString:
    FRONT_COLORS = [x.lower() for x in vars(Fore) if x != "RESET"]
    BACK_COLORS = [x.lower() for x in vars(Back) if x != "RESET"]
    STYLES = [x.lower() for x in vars(Style) if x != "RESET_ALL"]
    """
    cs = ColoredString()
    print(cs.red("abd"))
    print(cs.red_black("abc"))
    print(cs.red_black_bright("abc"))
    """

    def __init__(self):
        self.cache = {}

    def __getattr__(self, name):
        if name in self.cache:
            return self.cache[name]

        colors = name.split("_")
        back_color = None
        style = None

        if len(colors) == 1:
            front_color = colors[0]
        elif len(colors) == 2:
            front_color, back_color = colors
        else:
            front_color, back_color, style = colors[:3]

        front_colors = ColoredString.FRONT_COLORS
        if front_color not in front_colors:
            raise ValueError("available front colors:", front_colors)

        back_colors = ColoredString.BACK_COLORS
        if back_color and back_color not in back_colors:
            raise ValueError("available back colors:", back_colors)

        styles = ColoredString.STYLES
        if style and style not in styles:
            raise ValueError("available style:", styles)

        func = partial(
            message_in_color_front_and_back,
            getattr(Fore, front_color.upper()),
            getattr(Back, back_color.upper()) if back_color else None,
            getattr(Style, style.upper()) if style else None,
        )

        self.cache[name] = func
        return func


if __name__ == "__main__":
    cs = ColoredString()
    print(cs.red_black_bright("abc"))
    # print(cs.black_red("abc"))
    # print(cs.black_red_dim("abc"))
    # print(cs.funcs)
    # print(cs.red_yellow("abc"))
    # print(cs.red_black("cba"))
    # print(cs.black_green("abc"))



from datetime import datetime
from pathlib import Path
from typing import Any, Optional, Union
from uuid import UUID

import numpy as np
from PIL import Image

from ...core import HLDataModel, draw_annotations_on_image, resize_image

__all__ = ["DataFile"]


class DataFile(HLDataModel):
    file_id: Optional[UUID] = None
    content_type: str
    content: Any
    recorded_at: datetime = datetime.now()
    media_frame_index: int = 0
    original_source_url: Optional[str] = None

    @classmethod
    def from_image(
        cls,
        content: Union[np.ndarray, Image.Image, str, Path],
        file_id: UUID,
        media_frame_index: int = 0,
        original_source_url: Optional[str] = None,
    ):
        if isinstance(content, (str, Path)):
            content = Image.open(content)

        assert isinstance(content, (np.ndarray, Image.Image))

        return cls(
            file_id=file_id,
            content_type="image",
            content=content,
            media_frame_index=media_frame_index,
            original_source_url=original_source_url,
        )

    def get_id(self) -> UUID:
        return self.file_id

    def resize(self, width: int, height: int) -> "DataFile":
        assert self.content_type == "image"
        self.content = resize_image(self.content, width=width, height=height)
        return self

    def draw_annotations(self, annotations: "Annotations"):
        overlay = draw_annotations_on_image(self.content, annotations)
        return overlay



from logging import warning
from os import PathLike
from pathlib import Path
from typing import Optional, Tuple

import numpy as np
import pandas as pd
from shapely.geometry import Polygon
from shapely.wkt import loads as wkt_loads

from highlighter.agent.capabilities.sources import VideoFrameIterator
from highlighter.core.const import PIXEL_LOCATION_ATTRIBUTE_UUID
from highlighter.datasets.base_models import ImageRecord


def interpolate_pixel_locations_between_frames(
    annotations_df: pd.DataFrame,
    data_files_df: pd.DataFrame,
    frame_frac: Optional[float] = None,
    frame_count: Optional[int] = None,
    frame_save_dir: Optional[PathLike] = None,
    source_video_dir: Optional[PathLike] = None,
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    adf = annotations_df
    if "frame_id" not in adf.columns:
        try:
            frame_ids = adf.extra_fields.apply(lambda d: d["frame_id"])
            adf["frame_id"] = frame_ids
        except KeyError as _:
            raise ValueError(
                "Unable to determine frame_id. Expected a frame_id column or extra_fields with a 'frame_id' key"
            )

    sorted_adf = adf.sort_values(by="frame_id").set_index("frame_id").copy()

    # Convert WKT strings to Polygon objects only once
    sorted_adf.loc[sorted_adf.attribute_id == PIXEL_LOCATION_ATTRIBUTE_UUID, "value"] = sorted_adf[
        sorted_adf.attribute_id == PIXEL_LOCATION_ATTRIBUTE_UUID
    ].value.apply(lambda p: wkt_loads(p) if isinstance(p, str) else p)

    interpolated_rows = []

    for entity_id, grp in sorted_adf.groupby("entity_id"):
        loc_frames = grp[grp.attribute_id == PIXEL_LOCATION_ATTRIBUTE_UUID].index.to_numpy()
        loc_values = grp[grp.attribute_id == PIXEL_LOCATION_ATTRIBUTE_UUID].value.to_numpy()
        others_adf = grp[~(grp.attribute_id == PIXEL_LOCATION_ATTRIBUTE_UUID)]
        attrs = (
            others_adf.groupby("frame_id")
            .agg({"attribute_name": list, "attribute_id": list, "value": list})
            .iloc[0]
        )

        data_file_id = grp.data_file_id.iloc[0]
        dataset_id = grp.dataset_id.iloc[0]
        # Iterate over consecutive loc_frames
        for i in range(len(loc_frames) - 1):
            current_frame, next_frame = loc_frames[i], loc_frames[i + 1]
            current_bounds, next_bounds = loc_values[i].bounds, loc_values[i + 1].bounds
            frame_diff = next_frame - current_frame

            # Generate interpolated loc_values
            minX_vals = np.linspace(current_bounds[0], next_bounds[0], frame_diff + 1)[1:-1]
            minY_vals = np.linspace(current_bounds[1], next_bounds[1], frame_diff + 1)[1:-1]
            maxX_vals = np.linspace(current_bounds[2], next_bounds[2], frame_diff + 1)[1:-1]
            maxY_vals = np.linspace(current_bounds[3], next_bounds[3], frame_diff + 1)[1:-1]

            # Use itertools.chain to reduce append calls and improve performance
            for f, minX, minY, maxX, maxY in zip(
                range(current_frame + 1, next_frame), minX_vals, minY_vals, maxX_vals, maxY_vals
            ):
                interpolated_rows.append(
                    {
                        "frame_id": f,
                        "entity_id": entity_id,
                        "attribute_name": "pixel_location",
                        "attribute_id": PIXEL_LOCATION_ATTRIBUTE_UUID,
                        "value": Polygon(
                            [(minX, minY), (maxX, minY), (maxX, maxY), (minX, maxY), (minX, minY)]
                        ),
                        "data_file_id": f"{data_file_id}",
                        "dataset_id": int(dataset_id),
                    }
                )
                interpolated_rows.extend(
                    [
                        {
                            "frame_id": f,
                            "entity_id": entity_id,
                            "attribute_name": attribute_name,
                            "attribute_id": attribute_id,
                            "value": value,
                            "data_file_id": f"{data_file_id}",
                            "dataset_id": int(dataset_id),
                        }
                        for attribute_name, attribute_id, value in zip(*attrs.values)
                    ]
                )

    # Create DataFrame with interpolated rows and concatenate with the original DataFrame
    interpolated_adf = pd.concat([adf, pd.DataFrame(interpolated_rows)], ignore_index=True)
    # Keep the original_data_file_id, we'll use it later
    interpolated_adf["original_data_file_id"] = interpolated_adf.data_file_id.copy()
    # Append -{frame_id} to the data_file_id so each frame's data_file_id is unique
    interpolated_adf.loc[:, "data_file_id"] = (
        interpolated_adf["data_file_id"] + "-" + interpolated_adf["frame_id"].astype(str)
    )

    if (frame_frac is not None) or (frame_count is not None):
        interpolated_adf = interpolated_adf.sample(n=frame_count, frac=frame_frac)

    ddf = data_files_df

    def make_frame_data_files(grp, *, split):
        data_file_id = grp.name
        original_data_file_id = data_file_id.split("-")[0]
        original_data_file = ddf[ddf.data_file_id == original_data_file_id].iloc[0].to_dict()

        return pd.Series(
            ImageRecord(
                data_file_id=data_file_id,
                split=split,
                filename=f"{data_file_id}.jpg",
                width=original_data_file.get("width", None),
                height=original_data_file.get("height", None),
                assessment_id=original_data_file.get("assessment_id", None),
            ).model_dump()
        )

    # Preserve the original splits in the interpolated_ddf
    _split_ddfs = []
    for split in ddf.split.unique():
        # get all the data_files pre split for the original ds
        split_data_file_ids = ddf[ddf.split == split].data_file_id

        # filter out all the matching data_files from the interpolated annotations_df
        mask = interpolated_adf.data_file_id.map(lambda d: d.split("-")[0]).isin(split_data_file_ids)

        def fn(grp):
            return make_frame_data_files(grp, split=split)

        _split_ddfs.append(interpolated_adf[mask].groupby("data_file_id").apply(fn))
    interpolated_ddf = pd.concat(_split_ddfs, ignore_index=True)

    if frame_save_dir is not None:
        for original_data_file_id in ddf.data_file_id.unique():
            original_video_filename = ddf[ddf.data_file_id == original_data_file_id].iloc[0].filename
            if not original_video_filename.endswith(".mp4"):
                warning(f"Skipping {original_video_filename}, not a Video")
                continue

            source_video_path = (
                original_video_filename
                if source_video_dir is None
                else Path(source_video_dir) / original_video_filename
            )

            frame_ids = (
                interpolated_adf[interpolated_adf.original_data_file_id == original_data_file_id]
                .frame_id.unique()
                .tolist()
            )
            video_frame_iter = VideoFrameIterator(source_urls=[source_video_path])
            filename = f"{original_data_file_id}" + "-{}.jpg"
            extracted_frame_ids = video_frame_iter.extract_frames_by_id(
                frame_ids,
                frame_save_dir,
                filename,
            )
            interpolated_adf = interpolated_adf[interpolated_adf.frame_id.isin(extracted_frame_ids)]
            interpolated_ddf = interpolated_ddf[
                interpolated_ddf.data_file_id.isin(interpolated_ddf.data_file_id.unique())
            ]

    return interpolated_adf, interpolated_ddf



from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Any, List, Optional, Tuple, Union
from uuid import UUID

import yaml
from pydantic import BaseModel, ConfigDict, Field, ValidationError, field_validator

from ..core import GQLBaseModel
from .gql_client import HLClient

__all__ = [
    "EvaluationMetric",
    "EvaluationMetricCodeEnum",
    "EvaluationMetricResult",
    "create_evaluation_metric",
    "create_evaluation_metric_result",
    "find_or_create_evaluation_metric",
]


class EvaluationMetricCodeEnum(str, Enum):
    Dice = "Dice"
    mAP = "mAP"
    MaAD = "MaAD"
    MeAD = "MeAD"
    Other = "Other"
    AP = "AP"


class EvaluationMetric(GQLBaseModel):
    model_config = ConfigDict(use_enum_values=True)

    research_plan_id: int
    code: EvaluationMetricCodeEnum
    chart: Optional[str] = None
    description: Optional[str] = None
    iou: Optional[float] = None
    name: str
    object_class_uuid: Optional[Union[UUID, str]] = None
    weighted: Optional[bool] = False
    id: Optional[int] = None

    def dict(self, *args, **kwargs):
        d = super().model_dump(*args, **kwargs)
        if "object_class_uuid" in d:
            d["object_class_uuid"] = str(d["object_class_uuid"])
        return d


class EvaluationMetricResult(GQLBaseModel, extra="forbid"):
    research_plan_metric_id: int
    result: float
    # iso datetime str will be generated at instantiation
    # if not supplied manually.
    occured_at: Optional[Union[datetime, str]] = Field(..., default_factory=datetime.now)
    object_class_uuid: Optional[Union[UUID, str]] = None
    training_run_id: Optional[int] = None

    @field_validator("occured_at", mode="before")
    def set_timestamp(cls, v):
        if v is None:
            v = datetime.utcnow().isoformat()
        elif isinstance(v, str):
            v = datetime.fromisoformat(v).isoformat()
        elif isinstance(v, datetime):
            v = v.isoformat()
        else:
            raise ValidationError()
        return v

    @classmethod
    def from_yaml(cls, path: Union[Path, str]):
        path = Path(path)
        with path.open("r") as f:
            data = yaml.safe_load(f)
        return cls(**data)


def get_existing_evaluation_metrics(client: HLClient, evaluation_id: int):
    class QueryReturnType(GQLBaseModel):
        research_plan_metrics: List[EvaluationMetric]

    query_return_type: QueryReturnType = client.researchPlan(return_type=QueryReturnType, id=evaluation_id)

    return query_return_type.research_plan_metrics


def create_evaluation_metric(
    client: HLClient,
    evaluation_id: int,
    code: Union[EvaluationMetricCodeEnum, str],
    name: str,
    description: Optional[str] = None,
    iou: Optional[float] = None,
    weighted: Optional[bool] = False,
    object_class_uuid: Optional[Union[UUID, str]] = None,
) -> EvaluationMetric:
    # ToDo: Have the GQL accept s uuid not an id
    if isinstance(object_class_uuid, UUID):
        object_class_uuid = str(object_class_uuid)

    code = EvaluationMetricCodeEnum(code) if isinstance(code, str) else code

    class CreateResearchPlanMetricReturnType(GQLBaseModel):
        errors: Any = None
        research_plan_metric: Optional[EvaluationMetric] = None

    kwargs = EvaluationMetric(
        research_plan_id=evaluation_id,
        code=code,
        name=name,
        description=description,
        iou=iou,
        weighted=weighted,
        object_class_uuid=object_class_uuid,
    ).gql_dict()

    result = client.createResearchPlanMetric(
        return_type=CreateResearchPlanMetricReturnType, **kwargs
    ).research_plan_metric
    assert result is not None
    return result


def find_or_create_evaluation_metric(
    client: HLClient,
    evaluation_id: int,
    code: Union[EvaluationMetricCodeEnum, str],
    name: str,
    description: Optional[str] = None,
    iou: Optional[float] = None,
    weighted: Optional[bool] = False,
    object_class_uuid: Optional[Union[UUID, str]] = None,
) -> Tuple[EvaluationMetric, bool]:
    existing_evaluation_metrics = {r.name: r for r in get_existing_evaluation_metrics(client, evaluation_id)}

    if name in existing_evaluation_metrics:
        found = True
        result = existing_evaluation_metrics[name]
    else:
        found = False
        result = create_evaluation_metric(
            client,
            evaluation_id,
            code,
            name,
            description=description,
            iou=iou,
            weighted=weighted,
            object_class_uuid=object_class_uuid,
        )
    return result, found


def create_evaluation_metric_result(
    client: HLClient,
    evaluation_metric_id: int,
    result: Union[float, int],
    occured_at: Optional[Union[datetime, str]] = None,
    object_class_uuid: Optional[Union[str, UUID]] = None,
    training_run_id: Optional[int] = None,
):
    """Create an evaluation_metric if it does not exist, optionally
    create an evaluation_metric_result if result is not None
    """
    evaluation_metric_result = EvaluationMetricResult(
        research_plan_metric_id=evaluation_metric_id,
        result=result,
        occured_at=occured_at,
        object_class_uuid=object_class_uuid,
        training_run_id=training_run_id,
    )

    class CreateExperimentResultReturnType(BaseModel):
        errors: Any
        experimentResult: EvaluationMetricResult

    mutation_result: CreateExperimentResultReturnType = client.createExperimentResult(
        return_type=CreateExperimentResultReturnType,
        **evaluation_metric_result.gql_dict(),
    )
    return mutation_result.experimentResult



# To Do
# ~~~~~
# - Replace "Terminate" PipelineElement with proper Pipeline terminate approach

import tempfile
from typing import Tuple

import aiko_services as aiko

FRAME_DATA = "()"
GRACE_TIME = 60
PARAMETERS = {}

all = ["do_create_pipeline"]

class Terminate(aiko.PipelineElement):
    def __init__(self, context):
        context.set_protocol("terminate:0")
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream) -> Tuple[aiko.StreamEvent, dict]:
        self.logger.info(f"{self.my_id()}")
        aiko.process.terminate()  # TODO: Improve Aiko Services Process exit
        return aiko.StreamEvent.OKAY, {}

def do_create_pipeline(pipeline_definition_json):
    file = None
    with tempfile.NamedTemporaryFile(delete=True, mode="w") as file:
        file.write(pipeline_definition_json)
        file.flush()

        pipeline_definition =  \
            aiko.PipelineImpl.parse_pipeline_definition(file.name)

        pipeline = aiko.PipelineImpl.create_pipeline(
            file.name, pipeline_definition, name=None, graph_path=None,
            stream_id=None, parameters=PARAMETERS,
            frame_id=0, frame_data=FRAME_DATA,
            grace_time=GRACE_TIME, queue_response=None)

        pipeline.run(mqtt_connection_required=False)



from collections import defaultdict
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional, Tuple, Union
from uuid import UUID, uuid4

import numpy as np
import shapely.geometry as geom
from pydantic import Field
from shapely.wkt import loads as wkt_loads

from highlighter.client.base_models.base_models import SubmissionType
from highlighter.core.geometry import polygon_from_tlbr
from highlighter.core.hl_base_model import HasMany, HLModelMap

from ...core import (
    OBJECT_CLASS_ATTRIBUTE_UUID,
    PIXEL_LOCATION_ATTRIBUTE_UUID,
    HLDataModel,
)
from .annotation import Annotation
from .datum_source import DatumSource
from .observation import Observation

__all__ = ["Entity"]


@HasMany("annotations", target_cls="highlighter.client.Annotation", back_populates="entity")
@HasMany("global_observations", target_cls="highlighter.client.Observation", back_populates="entity")
class Entity(HLDataModel):
    id: UUID = Field(..., default_factory=uuid4)
    annotations: HLModelMap = Field(..., default_factory=lambda: HLModelMap(Annotation))
    global_observations: HLModelMap = Field(..., default_factory=lambda: HLModelMap(Observation))

    def get_annotations(self) -> List[Annotation]:
        return list(self.annotations.values())

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def make_entity(
        cls,
        entity_id: Optional[UUID] = None,
        location: Optional[geom.Polygon] = None,
        observations: List[Observation] = [],
        global_observations: List[Observation] = [],
    ):
        if entity_id is None:
            entity_id = uuid4()

        annotations = []
        if observations:
            annotations = [
                Annotation(
                    id=uuid4(),
                    entity_id=entity_id,
                    location=location,
                    observations=observations,
                    datum_source=observations[0].datum_source,
                )
            ]

        entity = cls(
            id=entity_id,
            annotations=annotations,
            global_observations=global_observations,
        )
        return entity

    def to_json(self):
        data = self.model_dump()
        data["id"] = str(data["id"])
        data["annotations"] = [a.to_json() for a in data["annotations"].values()]
        data["global_observations"] = [o.to_json() for o in data["global_observations"].values()]
        return data

    def serialize(self):
        # ToDo: Make this more general
        if self.annotations:
            annotations = [a.serialize() for a in self.annotations.values()]
        else:
            annotations = []

        if self.global_observations:
            global_observations = [o.serialize() for o in self.global_observations.values()]
        else:
            global_observations = []

        result = {
            "id": str(self.id),
            "annotations": annotations,
            "global_observations": global_observations,
        }
        return result

    def to_deprecated_observations(self) -> List[Observation]:
        """
        Convert to the deprecated "set of observations" representation
        where pixel locations are represented as observations rather than annotations
        """
        observations = []
        observations.extend(self.global_observations.values())
        for annotation in self.annotations:
            observations.extend(annotation.observations)
            if annotation.location is not None:
                observation = Observation.make_pixel_location_observation(
                    value=annotation.location,
                    confidence=annotation.datum_source.confidence,
                    occurred_at=datetime.now(timezone.utc),
                    frame_id=annotation.datum_source.frame_id,
                )
                observation.entity = annotation.entity
                observations.append(observation)
        return observations

    @staticmethod
    def entities_to_deprecated_observations(entities: Dict[UUID, "Entity"]) -> List[Observation]:
        observations = []
        for entity in entities.values():
            observations.extend(entity.to_deprecated_observations())
        return observations

    @staticmethod
    def entities_from_assessment(assessment: SubmissionType) -> Dict[UUID, "Entity"]:
        # Step 1: Group annotations and observations by entity, and further group observations by annotation
        grouped_entities = defaultdict(
            lambda: {"annotations": [], "observations": defaultdict(list), "global_observations": []}
        )
        for annotation in assessment.annotations:
            grouped_entities[UUID(annotation.entity_id)]["annotations"].append(annotation)
        for observation in assessment.entity_attribute_values:
            if observation.annotation_uuid is not None:
                grouped_entities[UUID(observation.entity_id)]["observations"][
                    observation.annotation_uuid
                ].append(observation)
            else:
                grouped_entities[UUID(observation.entity_id)]["global_observations"].append(observation)
        # Step 2: Convert Hl Web representations into our local representations
        entities = {}
        for entity_id, entity_data in grouped_entities.items():
            entity = Entity(id=entity_id)
            for annotation in entity_data["annotations"]:
                ann = Annotation(
                    entity_id=entity_id,
                    location=wkt_loads(annotation.location),
                    track_id=annotation.track_id,
                    data_file_id=annotation.data_file_id,
                    datum_source=DatumSource(
                        frame_id=annotation.frame_id,
                        confidence=annotation.confidence,
                    ),
                    correlation_id=annotation.correlation_id,
                )

                obs = Observation(
                    entity_id=entity_id,
                    attribute_id=OBJECT_CLASS_ATTRIBUTE_UUID,
                    value=annotation.object_class.uuid,
                    occurred_at=datetime.now(
                        timezone.utc
                    ),  # TODO store with annotation in hl web or infer from source data file
                    datum_source=DatumSource(
                        frame_id=annotation.frame_id,
                        confidence=annotation.confidence,
                    ),
                )
                ann.observations.add(obs)
                entity.annotations.add(ann)

                for eav in entity_data["observations"][annotation.uuid]:
                    obs = Observation(
                        entity_id=entity_id,
                        attribute_id=eav.entity_attribute_id,
                        value=(
                            eav.value
                            if eav.value is not None
                            else eav.related_entity_id or eav.file_uuid or eav.entity_attribute_enum.id
                        ),
                        occurred_at=eav.occurred_at,
                        datum_source=eav.entity_datum_source,
                    )
                    ann.observations.add(obs)

            for eav in entity_data["global_observations"]:
                entity.global_observations.add(
                    Observation(
                        entity_id=entity_id,
                        attribute_id=eav.entity_attribute_id,
                        value=(
                            eav.value
                            if eav.value is not None
                            else eav.related_entity_id or eav.file_uuid or eav.entity_attribute_enum.id
                        ),
                        occurred_at=eav.occurred_at,
                        datum_source=eav.entity_datum_source,
                    )
                )
            entities[entity_id] = entity
        return entities

    @staticmethod
    def frame_indexed_entities_from_avro(
        avro_entities, data_file_id: UUID
    ) -> List[Tuple[int, Dict[UUID, "Entity"]]]:
        """See Avro schema at highlighter.entity_avro_schema"""
        frame_indexed_entities = defaultdict(dict)  # Outer index is frame ID, inner index is entity ID
        for entity in avro_entities:
            # TODO handle embeddings
            # TODO handle eavts
            for track in entity.tracks:
                for detection in track.detections:
                    observations = [
                        Observation(
                            entity_id=entity.id,
                            attribute_id=OBJECT_CLASS_ATTRIBUTE_UUID,
                            value=entity.object_class,
                            occurred_at=datetime.now(timezone.utc),  # TODO change to correct occurred_at
                            datum_source=DatumSource(
                                confidence=1.0,
                                frame_id=detection.frame_id,
                            ),
                        )
                    ]
                    annotations = [
                        Annotation(
                            id=uuid4(),
                            entity_id=entity.id,
                            location=polygon_from_tlbr(detection.bounds),
                            track_id=track.track_id,
                            observations=observations,
                            data_file_id=data_file_id,
                            datum_source=DatumSource(
                                confidence=1.0,
                                frame_id=detection.frame_id,
                            ),
                        )
                    ]
                    global_observations = []
                    frame_indexed_entities[detection.frame_id][entity.id] = Entity(
                        id=entity.id,
                        annotations=annotations,
                        global_observations=global_observations,
                    )
        raise NotImplementedError("This implementation is a sketch, don't use without adding tests")
        return sorted(frame_indexed_entities.items(), key=lambda kv: kv[0])

    @staticmethod
    def entities_from_deprecated_eavts(eavts: List["EAVT"]) -> Dict[UUID, "Entity"]:
        entities = {}
        if len(eavts) > 0:
            grouped = defaultdict(list)
            for eavt in eavts:
                grouped[eavt.entity_id].append(eavt)
            for group in grouped.values():
                entity_id = group[0].entity_id
                location_eavts = [e for e in group if e.attribute_id == PIXEL_LOCATION_ATTRIBUTE_UUID]
                if len(location_eavts) == 0:
                    entities[entity_id] = Entity(id=entity_id)

                    for eavt in group:
                        obs_id = uuid4()
                        obs = Observation.from_deprecated_eavt(
                            eavt,
                            obs_id,
                        )

                        entities[entity_id].global_observations.add(obs)

                elif len(location_eavts) == 1:
                    location_eavt = location_eavts[0]
                    entity = Entity(id=entity_id)
                    annotation = Annotation(
                        id=uuid4(),
                        # The EAVT 'value' is a PixelLocationAttributeValue
                        # which has a shapely geometry as its 'value'
                        location=location_eavt.value.value,
                        datum_source=location_eavt.datum_source,
                    )

                    for eavt in group:
                        if eavt.attribute_id != PIXEL_LOCATION_ATTRIBUTE_UUID:
                            obs_id = uuid4()
                            obs = Observation.from_deprecated_eavt(
                                eavt,
                                obs_id,
                            )
                            annotation.observations.add(obs)

                    entity.annotations.add(annotation)
                    entities[entity_id] = entity
                else:
                    raise ValueError(
                        f"Can't handle {len(location_eavts)} pixel locations for a single entity"
                    )
        return entities

    def crop_annotations(
        self,
        crop_args: "CropArgs",
        crop_annotations_with: Optional[Union[UUID, Tuple[UUID, Any]]] = None,
    ) -> List[np.ndarray]:
        crops = []
        if crop_annotations_with is None:
            for a in self.annotations:
                crops.append(a.crop(crop_args))
        else:

            if isinstance(crop_annotations_with, UUID):
                value = None
                attribute_id = crop_annotations_with
            else:
                attribute_id, value = crop_annotations_with

            for a in self.annotations:
                if a.has_observation(attribute_id, value=value):
                    crops.append(a.crop(crop_args))

        return crops

    @staticmethod
    def crop_entity_annotations(
        entities: List["Entity"],
        crop_args: "CropArgs",
        crop_annotations_with: Optional[Union[UUID, Tuple[UUID, Any]]] = None,
    ) -> List[np.ndarray]:
        crops = []
        for e in entities:
            crops.extend(e.crop_annotations(crop_args, crop_annotations_with=crop_annotations_with))
        return crops



import os
from pathlib import Path
from typing import Optional, Union

import numpy as np
import onnxruntime as ort
from pydantic import HttpUrl

from highlighter.client.io import download_bytes, get_cache_path
from highlighter.core import get_default_logger

HL_GPU_DEVICE_ID = os.environ.get("HL_GPU_DEVICE_ID", 0)
HL_MODEL_HEAD_INDEX = os.environ.get("HL_MODEL_HEAD_INDEX", 0)
LOGGER = get_default_logger("OnnxPredictor")
ort.set_default_logger_severity(os.environ.get("ONNX_RUNTIME_LOGLEVEL", 3))


def is_cuda_available():
    providers = ort.get_available_providers()
    return "CUDAExecutionProvider" in providers


def is_url(s):
    try:
        url = HttpUrl(s).unicode_string()  # pyright: ignore
        return url
    except Exception:
        return False


class OnnxPredictor:
    """Loads an onnx model and does prediction

    Set log level:
       set the ONNX_RUNTIME_LOGLEVEL environment variable

    Set gpu id
       set the HL_GPU_DEVICE_ID environment variable

    Select a different head when reading TrainingRunInterface.model_outputs
       set the HL_MODEL_HEAD_INDEX environment variable
    """

    def __init__(
        self,
        onnx_file: Union[str, Path],
        device_id=HL_GPU_DEVICE_ID,
        ort_custom_op_path: Optional[str] = None,
    ):
        session_options = ort.SessionOptions()
        # register custom op for onnxruntime
        if ort_custom_op_path not in (None, ""):
            session_options.register_custom_ops_library(str(ort_custom_op_path))

        providers = ["CPUExecutionProvider"]
        provider_options = [{}]
        self.is_cuda_available = is_cuda_available()
        if self.is_cuda_available:
            providers.insert(0, "CUDAExecutionProvider")
            provider_options.insert(0, {"device_id": device_id})
        LOGGER.debug(f"is_cuda_available={self.is_cuda_available}")
        LOGGER.debug(f"provider_options={provider_options}")

        if url := is_url(onnx_file):
            onnx_file = get_cache_path(url)
            download_bytes(url, save_path=onnx_file, check_cached=True)

        self.sess = ort.InferenceSession(
            str(onnx_file),
            sess_options=session_options,
            providers=providers,
            provider_options=provider_options,
        )

        self.io_binding = self.sess.io_binding()

        _inputs = self.sess.get_inputs()
        if len(_inputs) > 1:
            raise ValueError(f"Expected only 1 input name for onnx-runtime session, got: {_inputs}")
        self.input_name = _inputs[0].name
        input_shape = _inputs[0].shape

        # 3D input probably indicates an ONNX model that expects a single image
        if len(input_shape) == 3:
            raise ValueError(
                f"Onnx models must have batch dimension. This Onnx model has an input of shape {input_shape}"
            )
        self.batch_size = input_shape[0]
        if isinstance(self.batch_size, int) and self.batch_size != 1:
            raise ValueError(
                f"Batch size must be the integer 1 or a symbolic value to indicate dynamic match size, got: {type(self.batch_size).__name__}: {self.batch_size}"
            )

        self.output_names = [_.name for _ in self.sess.get_outputs()]
        LOGGER.debug(f"onnxruntime session instantiated, input_shape {input_shape}")

    def predict(self, input_batch):
        if self.batch_size == 1:
            outputs = []
            for single_input in input_batch:
                single_input = np.expand_dims(single_input, axis=0)
                outputs.append(self.predict_batch(single_input)[0])

        elif not isinstance(self.batch_size, int):  # Symbolic value indicates dynamic batch size
            outputs = self.predict_batch(input_batch)
        else:
            raise ValueError(f"Unhandled batch size: {self.batch_size}")
        return outputs

    def predict_batch(self, input_batch):
        self.io_binding.bind_cpu_input(  # No-op if gpu is not available
            self.input_name,
            input_batch,
        )

        for name in self.output_names:
            self.io_binding.bind_output(name)

        # run session to get outputs
        self.sess.run_with_iobinding(self.io_binding)
        ort_outputs = self.io_binding.copy_outputs_to_cpu()  # No copy is performed if gpu is not available
        return ort_outputs



import warnings
from datetime import datetime
from pathlib import Path
from typing import Any, List, Optional, Union

from ..core import GQLBaseModel, paginate
from .aws_s3 import upload_file_to_s3
from .base_models import Image as ImageType
from .base_models import PageInfo
from .gql_client import HLClient
from .presigned_url import get_presigned_url, get_presigned_urls

__all__ = [
    "get_presigned_data_file",
    "get_presigned_data_files",
    "get_data_files",
    "create_data_file",
]


def get_data_files(
    client,
    data_file_ids: Optional[List[int]] = None,
    data_source_id: Optional[List[int]] = None,
):
    class ImageTypeConnection(GQLBaseModel):
        page_info: PageInfo
        nodes: List[ImageType]

    kwargs = {"id": data_file_ids, "dataSourceId": data_source_id}
    kwargs = {k: v for k, v in kwargs.items() if v is not None}

    return paginate(client.imageConnection, ImageTypeConnection, **kwargs)


def get_presigned_data_file(
    client: HLClient,
    id: int,
):
    warnings.warn("get_presigned_data_file is deprecated use get_presigned_url")
    return get_presigned_url(
        client,
        id,
    )


def get_presigned_data_files(
    client,
    ids: List[int],
):
    warnings.warn("get_presigned_data_files is deprecated use get_presigned_urls")
    return get_presigned_urls(
        client,
        ids,
    )


def create_data_file(
    client: HLClient,
    data_file_path: Union[str, Path],
    data_source_id: int,
    site_id: Optional[str] = None,
    observed_timezone: Optional[str] = None,
    recorded_at: Optional[str] = None,
    metadata: str = "{}",
    uuid: Optional[str] = None,
    multipart_filesize: Optional[str] = None,
) -> ImageType:
    data_file_path = Path(data_file_path)
    if not data_file_path.exists():
        raise FileNotFoundError(f"{data_file_path}")

    file_data = upload_file_to_s3(client, str(data_file_path), multipart_filesize=multipart_filesize)

    if recorded_at is None:
        recorded_at = datetime.now().isoformat()

    class CreateImageResponse(GQLBaseModel):
        image: Optional[ImageType] = None
        errors: Any = None

    create_data_file_response = client.create_image(
        return_type=CreateImageResponse,
        dataSourceId=data_source_id,
        originalSourceUrl=str(data_file_path),
        fileData=file_data,
        siteId=site_id,
        observedTimezone=observed_timezone,
        recordedAt=recorded_at,
        metadata=metadata,
        uuid=uuid,
        contentType="image",
    )

    if create_data_file_response.errors:
        raise ValueError(". ".join(create_data_file_response.errors))
    return create_data_file_response.image



# Notes
# ~~~~~
# This is a low-level utility used by the AikoServices framework, so be
# careful when making static dependencies on the AikoServices framework.
#
# Python's "logging" package is NOT intuitive :(
# https://stackoverflow.com/questions/43109355/logging-setlevel-is-being-ignored
# https://docs.python.org/3/library/logging.html
# https://docs.python.org/3/library/logging.html#formatter-objects
# https://docs.python.org/3/library/stdtypes.html#printf-style-string-formatting
#
# Environment variables
# ~~~~~~~~~~~~~~~~~~~~~
# AIKO_LOG_LEVEL:         ERROR, WARNING, INFO, DEBUG
# AIKO_LOG_MQTT:          false: Console, true: MQTT, all: Both Console / MQTT
# AIKO_LOG_REPEAT_PERIOD: Period for repeated log messages, default 6 seconds
# AIKO_LOG_REPEAT_DEPTH:  To be implemented
#
# Typical usage: Aiko and LoggingHandlerMQTT
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# AIKO_LOG_LEVEL=INFO python                      # MQTT logging (default)
# AIKO_LOG_LEVEL=INFO AIKO_LOG_MQTT=false python  # Console logging
#   from aiko_services.main import *
#   _LOGGER = aiko.logger(__name__)
#   _LOGGER.info("Informative message")
#   aiko.process.run(True)  # Required for MQTT based logging
#
# Direct usage: Standalone Logger
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# AIKO_LOG_LEVEL=DEBUG python  # AIKO_LOG_LEVEL=INFO (default)
#   from aiko_services.main.utilities import get_logger
#   _LOGGER = get_logger(__name__)
#   _LOGGER.debug("Diagnostic message")
#
# Usage: Change logging level on-the-fly using ECProducer
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# from aiko_services.main import *
# from aiko_services.main.utilities import *
# _LOGGER = aiko.logger(__name__)
#
# def ec_producer_change_handler(command, item_name, item_value):
#     if item_name == "log_level":
#         _LOGGER.setLevel(log_level_real(item_value))
#
# state = {"log_level": get_log_level_name(_LOGGER)}
# ec_producer = ECProducer(service, state)
# ec_producer.add_handler(ec_producer_change_handler)
# _LOGGER.info("Informative message")
# aiko.process.run(True)  # Required for ECProducer and MQTT based logging
#
# To Do: Logger
# ~~~~~~~~~~~~~
# * Implement set_log_level(name, level) and keep a dictionary of all loggers,
#   so that an individual, a list or all loggers can be changed at once
#
# - Based on current log_level, select _LOG_FORMAT/_DEBUG on-the-fly
# - Allow AIKO_LOG_MQTT value to be changed on-the-fly (via ECProducer)
#
# - Implement logging to a file (break into chunks by date/time or size)
#   - Set log file path and logging level via command line arguments
# - Turn "logger" into a Python class, nothing global !
#
# To Do: LoggingHandlerMQTT
# ~~~~~~~~~~~~~~~~~~~~~~~~~
# * Add support for AIKO_LOG_REPEAT_DEPTH to handle multiple repeated messages
#
# * Add support for ConnectionState.TRANSPORT (MQTT) connection updates

from collections import deque
import logging
from logging.config import dictConfig
import os
import sys
import time
from typing import Any

from aiko_services.main.utilities import *

__all__ = [
    "DEBUG", "get_level_name", "get_logger", "LoggingHandlerMQTT",
    "log_level_all", "log_level_real", "print_error"
]

DEBUG = logging.DEBUG
DEFAULT_LOG_REPEAT_PERIOD = "6"  # seconds for "AIKO_LOG_REPEAT_PERIOD"

_RING_BUFFER_SIZE = 128  # Maximum log messages to store before MQTT available

_LEVEL_NAMES = {
    0: "LOG_LEVEL_NOTSET",        #  0
    logging.DEBUG: "DEBUG",       # 10
    logging.INFO: "INFO",         # 20
    logging.WARNING: "WARNING",   # 30
    logging.ERROR: "ERROR",       # 40
    logging.CRITICAL: "CRITICAL"  # 50
}

_LOG_FORMAT = "%(asctime)s.%(msecs)03d %(levelname) 8s"
_LOG_FORMAT_DEBUG = f"{_LOG_FORMAT}s %(module)s.%(funcName)s() %(message)s"
_LOG_FORMAT_DEFAULT = f"{_LOG_FORMAT} %(name)18s %(message)s"
_LOG_FORMAT_DATETIME = "%Y-%m-%d_%H:%M:%S"

def get_log_level_name(logger):
    if logger.level in _LEVEL_NAMES:
        level_name = _LEVEL_NAMES[logger.level]
    else:
        level_name = str(logger.level)
    return level_name

def get_logger(name: str, log_level=None, logging_handler=None) -> Any:
#   logging.basicConfig(filename="aiko.log")  # TODO: Implement logging to file

    name = name.rpartition('.')[-1].upper()
#   print(f"Create logger {name}: {log_level}, {logging_handler}")

    if log_level is None:
        log_level = os.environ.get("AIKO_LOG_LEVEL", logging.INFO)
    if log_level == "":
        log_level = logging.INFO

    if logging_handler is None:
        logging_handler = logging.StreamHandler()

    format = _LOG_FORMAT_DEFAULT  # _LOG_FORMAT_DEBUG
    formatter = logging.Formatter(format, datefmt=_LOG_FORMAT_DATETIME)
    logging_handler.setFormatter(formatter)
    logger = logging.getLogger(name)
    logger.addHandler(logging_handler)
    logger.setLevel(log_level)
    return logger

def log_level_all(log_level):
    return log_level.upper().endswith("_ALL")

def log_level_real(log_level):  # Make uppercase and strip "_ALL" suffix
    log_level = log_level.upper()
    log_level = log_level[:-4] if log_level_all(log_level) else log_level
    return log_level

def print_error(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)

# -----------------------------------------------------------------------------

from aiko_services.main.connection import ConnectionState

class LoggingHandlerMQTT(logging.Handler):
    def __init__(
        self, aiko, topic, option="all", ring_buffer_size=_RING_BUFFER_SIZE):

        super().__init__()
        self.aiko = aiko
        self.console_flag = option == "all"
        self.topic = topic

        self.repeated_period = os.environ.get(
            "AIKO_LOG_REPEAT_PERIOD", DEFAULT_LOG_REPEAT_PERIOD)
        if self.repeated_period.isnumeric():
            self.repeated_period = int(self.repeated_period)
        else:
            self.repeated_period = 0

        self.ready = False
        self.record_previous = None
        self.record_repeat_count = 0
        self.record_time = 0
        self.ring_buffer = deque(maxlen=ring_buffer_size)
        aiko.connection.add_handler(self._connection_state_handler)

    def _connection_state_handler(self, connection, connection_state):
        if connection.is_connected(ConnectionState.TRANSPORT):
            self.ready = True
            self._publish_ring_buffer()
        else:
            self.ready = False

    def __del__(self):
        try:
            pass  # TODO: release resources
        except:
            pass
        try:
            pass  # TODO: self.???? = None
        except:
            pass

    def emit(self, record):  # record: logging.LogRecord has lots of details
        try:
            repeated, repeated_message = self._repeated_record(record)
            if repeated:
                if repeated_message:
                    record.msg = repeated_message
                else:
                    return

            payload_out = self.format(record)
            if self.console_flag:
                try:
                    print(payload_out)  # TODO: For shell pipes ... flush=True
                except BrokenPipeError:
                    pass

            if self.ready:
                self._publish_ring_buffer()
                self.aiko.message.publish(self.topic, payload_out)
            else:
                self.ring_buffer.append(payload_out)
            self.flush()
    # TODO: Catch specific Exceptions, don't hide other problems !
        except Exception as exception:  # TODO: When does this occur ?
            self.handleError(record)  # TODO: Start buffering log records

    def _publish_ring_buffer(self):
        while len(self.ring_buffer):
            payload_out = self.ring_buffer.popleft()
            self.aiko.message.publish(self.topic, payload_out)

    def _repeated_record(self, record):
        message = None
        repeated = self.record_previous and  \
                   self.record_previous.msg == record.msg and  \
                   self.record_previous.name == record.name and  \
                   self.record_previous.levelname == record.levelname
        if self.repeated_period > 0 and repeated:
            self.record_repeat_count += 1
            if time.time() > self.record_time + self.repeated_period:
                message = f"Repeated message count: {self.record_repeat_count}"
                self.record_repeat_count = 0
                self.record_time = time.time()
        else:
            repeated = False
            self.record_previous = record
            self.record_repeat_count = 0
            self.record_time = time.time()
        return repeated, message

# -----------------------------------------------------------------------------



import importlib
import json
from enum import Enum
from pathlib import Path


class TrainerType(str, Enum):
    YOLO_DET = "yolo-det"
    YOLO_SEG = "yolo-seg"
    YOLO_CLS = "yolo-cls"
    # ToDo: Add when the sdk can generate the correct dataset formats
    # YOLO_POSE = "yolo-det"
    # YOLO_OOB = "yolo-det"


class DIRS:

    @staticmethod
    def hl_cache(scaffold_dir: Path) -> Path:
        return scaffold_dir / "highlighter"

    @staticmethod
    def hl_context_json(scaffold_dir: Path) -> Path:
        return DIRS.hl_cache(scaffold_dir) / "context.json"

    @staticmethod
    def hl_training_run_cache_dir(scaffold_dir: Path, training_run_id: str) -> Path:
        """This is for storing non-user facing dataset stuff"""
        return DIRS.hl_cache(scaffold_dir) / "ml_training" / training_run_id

    @staticmethod
    def hl_training_config(scaffold_dir: Path, training_run_id: str) -> Path:
        """This is for storing non-user facing dataset stuff"""
        return DIRS.hl_training_run_cache_dir(scaffold_dir, training_run_id) / "training_config.json"

    @staticmethod
    def hl_training_run_cache_datasets_dir(scaffold_dir: Path, training_run_id: str) -> Path:
        """This is for storing non-user facing dataset stuff"""
        return DIRS.hl_cache(scaffold_dir) / "ml_training" / training_run_id / "datasets"

    @staticmethod
    def training_run_dir(scaffold_dir: Path, training_run_id: str) -> Path:
        """This is for storing user facing dataset stuff"""
        return scaffold_dir / "ml_training" / training_run_id


def load_hl_context(ctx_pth: Path):
    with ctx_pth.open("r") as f:
        ctx = json.load(f)
    return ctx


def append_hl_scaffold_context(scaffold_dir, items):
    ctx_pth = DIRS.hl_context_json(scaffold_dir)
    ctx = load_hl_context(ctx_pth)

    for k, v in items.items():
        if (k in ctx) and (ctx[k] != v):
            raise ValueError(f"Cached value {k}: {ctx[k]} conflicts with new value {v}")
        ctx[k] = v

    with ctx_pth.open("w") as f:
        json.dump(ctx, f, indent=2)


def load_trainer_module(training_run_dir):
    trainer_module_path = training_run_dir / "trainer.py"
    module_name = "trainer"
    spec = importlib.util.spec_from_file_location(module_name, trainer_module_path)
    trainer = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(trainer)
    return trainer


def ask_to_overwrite(dir, terminate=False):
    overwrite = input(f"Do you want to overwrite '{dir}'? (y/n):").lower() in ("y", "yes")
    if overwrite:
        return True
    if terminate:
        raise SystemExit(f"Terminating, {dir} exists.")
    else:
        return False



from asciimatics.widgets import Layout, MultiColumnListBox, Widget

import aiko_services as aiko  # designed to be external to the main framework

__all__ = []

class RegistrarFrame(aiko.ServiceFrame):
    def __init__(self, screen, dashboard):
        super(RegistrarFrame, self).__init__(
            screen, dashboard, name="registrar_frame")

        self.services_cache = aiko.services_cache_create_singleton(
            aiko.process, True, history_limit=screen.height)

        self._registrar_widget = MultiColumnListBox(
            screen.height * 1 // 2, ["<0"], options=[],
            titles=["Registrar: Discovered Services topic paths"]
        )
        layout_0 = Layout([1])
        self.add_layout(layout_0)
        layout_0.add_widget(self._registrar_widget)
        self.log_ui = aiko.LogUI(self)
        self.fix()  # Prepare asciimatics_Frame for use

    def _service_frame_start(self, service, service_ec_consumer):
        self.log_ui._service_frame_start(service, service_ec_consumer)

    def _service_frame_stop(self, service):
        self.log_ui._service_frame_stop(service)

    def _update(self, frame_no):
        super(RegistrarFrame, self)._update(frame_no)

        services = self.services_cache.get_services().copy()
        services_formatted = []
        for service in services:
            topic_path = str(aiko.ServiceTopicPath.parse(service[0]))
            protocol = service[2]  # self._short_name(service[2])
            services_formatted.append(
                (topic_path, service[1], service[4], protocol, service[3]))
        self._registrar_widget.options = [
            (service_info, row_index)
            for row_index, service_info in enumerate(services_formatted)
        ]

        self.log_ui._update(frame_no)

# plugin key can be either the Service "name" or "protocol"

plugins = {
    "registrar": RegistrarFrame
}



# Usage
# ~~~~~
# pytest [-s] unit/test_context.py
# pytest [-s] unit/test_context.py::test_actor_args
#
# test_actor_args()
# - Create initial "actor_args" for creating Actors via composition
#
# To Do
# ~~~~~
# - Check all "context.py" init_args convenience functions
# - Check using all "init_args" for creating Instances via composition

import aiko_services as aiko

ACTOR_NAME = "Alice"
TRANSPORT = "mqtt"

# init_args{
#     "context": ContextService(
#         name="Name",
#         implementations=None,
#         parameters={},
#         protocol="*",
#         tags=[],
#         transport="mqtt")
# }

def actor_args(name):
    return aiko.actor_args(name)

def test_actor_args():
    init_args = actor_args(ACTOR_NAME)
    assert init_args is not None
    assert "context" in init_args

    context = init_args["context"]
    assert isinstance(context, aiko.ContextService)
    assert context.name == ACTOR_NAME
    assert context.implementations is None
    assert context.parameters == {}
    assert context.protocol == "*"
    assert context.tags == []
    assert context.transport == TRANSPORT



#!/usr/bin/env python3
#
# Usage
# ~~~~~
# ./parser.py  # parse() <--> generate() tests for a range of examples
#
# For all operations, the parser() and generate() are each other's inverse
# - parse(generate(atom, list))     --> atom, list
# - generate(parse("s_expression")) --> "s_expression"
#
# Parse lists of symbols recursively
#
# - parse("")
# - parse("()")
# - parse("(c)")
# - parse("(c p1 p2)")
# - parse("(add topic protocol owner (a=b c=d))")
# - parse("('aloha honua')") or parse('("aloha honua")')  # quoted strings
#
# Parse dictionaries with keyword / value pairs
#
# - parse("(a (b: ''))")      --> ("a", [{'b': ""}])
# - parse("(a: 1 b: 2)")      --> {"a": 1, "b": 2}
# - parse("(a: (b c))")       --> {"a": ["b" "c"]}
# - parse("(a: (b: 1 c: 2))") --> {"a": {"b": "1", "c": "2"}}
# - parse("(a: 1 b)")         Illegal: Dictionary and positional parameters
# - parse("(a: 1 (b c) 2)")   Illegal: (b c) should be a dictionary keyword
#
# Parse canonical S-Expressions: binary symbols are prefixed with their length
#   https://en.wikipedia.org/wiki/Canonical_S-expressions
#
# - parse("a 0: b")           --> ["a None b"]
# - parse("3:a b")            --> ["a b"]
# - parse("3:a b 3:c d")      --> ["a b", "c d"]
#
# Doesn't parse plain symbols, i.e anything not in a (list)
# Doesn't parse quoted tokens, e.g parse("(c '(this_is_not_a_list)')") fails
#
# To Do
# ~~~~~
# - Consolidate generate(atom, list) into generate(list)
# - Consolidate parser result "(atom list)" into "list"
#   - Ensure that any S-Expression can be generated and parsed, not only "list"
#
# - Provide proper unit tests !
#
# - Change parse() to simply return the complete tree and ...
#     create new parse_payload() function that returns (command, parameters)
# - Change generate() to simply create the complete S-Expression and ...
#     create new generate_payload() function that returns (command, parameters)
#
# - Implement AVRO schema and JSON parsing, refactor "pipeline.py" ?
# - Incorporate Python module "sexpdata"
#   - https://sexpdata.readthedocs.io/en/latest
#   - https://github.com/jd-boyd/sexpdata
# - Incorporate Python module "hy" and "hyrule"
#
# - Review all places where parse() and generate() are and should be used ...
#   - parse() ...
#       actor.py:     _topic_in_handler()
#       lifecycle.py: _lcm_topic_contol_handler()
#       process.py:    on_registrar()
#       registrar.py: _service_state_handler()
#                     _topic_in_handler()
#       share.py:     _procedure_handler()
#                     _consumer_handler()
#                      registrar_share_handler()
#                      registrar_out_handler()
#       storage.py:   _topic_in_handler()
#                      do_request()
#   - generate() ...
#       Manually "generated" S-Expression payloads everywhere !
#       process.py:   _add_service_to_registrar(): should use generate()
#                     _remove_service_to_registrar(): should use generate()
#       recorder.py:  _recorder_handler(): should use generate()
#       share.py:     _dictionary_to_commands()
#                     _update_consumers(): should use generate()
#       transport/transport_mqtt.py: make_proxy_mqtt()

import re
import sys
from typing import Any, Dict, List, Tuple, Union

__all__ = ["generate", "parse", "parse_float", "parse_int", "parse_number"]

def generate(command: str, parameters: Union[Dict, List, Tuple]) -> str:
    if isinstance(parameters, dict):
        parameters = generate_dict_to_list(parameters)
    elif isinstance(parameters, tuple):
        parameters = list(parameters)
    expression = [command] + parameters
    return generate_s_expression(expression)

def generate_dict_to_list(expression: Dict) -> list:
    result = []
    for keyword, value in expression.items():
        result.append(f"{keyword}:")
        result.append(value)
    return result

RE_DELIMITERS = re.compile(r"^\d+:|[\s()]")

def generate_s_expression(expression: List) -> str:
    character = ""
    payload = "("
    for element in expression:
        if isinstance(element, str):
            if RE_DELIMITERS.search(element):
                element = f"{len(element)}:{element}"
        if isinstance(element, dict):
            element = generate_dict_to_list(element)
        if isinstance(element, list) or isinstance(element, tuple):
            element = generate_s_expression(element)
        if element == "":
            character = ' ""'
        if element is None:
            element = "0:"
        payload = f"{payload}{character}{element}"
        character = " "
    payload = f"{payload})"
    return payload

RE_CANONICAL_SYMBOL = re.compile(r"^(\d+):(.+)", re.DOTALL)  # length:data
RE_STRING = re.compile(r"""(['"])(.*?)\1""")                 # "text" or 'text'

def parse(payload: str, car_cdr=True, dictionaries_flag=True):
    result = []
    token = ""
    i = 0
    while i < len(payload):
        if not token:
            match = RE_CANONICAL_SYMBOL.match(payload[i:])
            if match:
                token_length = match.group(1)
                if int(token_length) == 0:
                    token = None
                else:
                    token = match.group(2)[:int(token_length)]
                result.append(token)
                token = ""
                i += len(token_length) + 1 + int(token_length)
                continue

            match = RE_STRING.match(payload[i:])
            if match:
                result.append(match.group(2))  # token
                i += len(match.group(0))       # token_length
                continue

        c = payload[i]
        if c == "(":
            sublist, j = parse(payload[i+1:])
            i += j
            result.append(sublist)
        elif c == ")":
            if token:
                result.append(token)
                token = ""
            return result, i+1
        elif c in [" ", "\t", "\n"]:
            if token:
                result.append(token)
                token = ""
        else:
            token += c
        i += 1
    if token:
        result.append(token)

    car = ""
    cdr = []
    if result:
        if isinstance(result[0], str):
            car = result[0]
        else:
            try:
                car = result[0][0]
                cdr = result[0][1:]
            except IndexError:
                pass

    if dictionaries_flag:
        cdr = parse_list_to_dict(cdr)

    if car_cdr:
        return car, cdr
    else:
        cdr.insert(0, car)
        return cdr

def parse_float(payload: str, default: float=0.0) -> float:
    try:
        result = float(payload)
    except ValueError:
        result = default
    return result

def parse_int(payload: str, default: int=0) -> int:
    try:
        result = int(payload)
    except ValueError:
        result = default
    return result

def parse_list_to_dict(tree: Any) -> Union[list, dict]:
    parse_error = "Error parsing S-Expression dictionary starting at keyword"
    result = tree
    if isinstance(tree, list) and len(tree):
        car = tree[0]
        if isinstance(car, str) and len(car) and car[-1] == ":":
            if len(tree) % 2 != 0:
                raise ValueError(f'{parse_error} "{car}", must have pairs of keywords and values')
            result = {}
            for i in range(len(tree) // 2):
                if not isinstance(tree[i*2], str):
                    raise ValueError(f'{parse_error} "{tree[i*2]}", keyword must be a string')
                if len(tree[i*2]) and tree[i*2][-1] != ":":
                    raise ValueError(f'{parse_error} "{tree[i*2]}", keyword must end with ":" character')
                keyword = tree[i*2][:-1]
                value = parse_list_to_dict(tree[i*2+1])
                result[keyword] = value
        else:
            result = [parse_list_to_dict(element) for element in tree]
    return result

def parse_number(payload: str, default: int=0):
    try:
        result = int(payload)
    except (TypeError, ValueError):
        try:
            result = float(payload)
        except (TypeError, ValueError):
            result = default
    return result

def main():
    payloads = [
    #   "abc",                      # Fails: generate() returns list !
    #   "abc def",                  # Fails: parse() only handles lists
        "(a 0: b)",                 # List containing None (encoded as 0:)
        "(a b ())",                 # List containing empty list
        "(a b (c d))",              # List containing list
        "(a b (c d) (e f (g h)))",  # List containing lists
        "(a b: 1 c: 2)",            # Dictionary
        "(a b: 1 c: (d e))",        # Dictionary containing list
        "(a b: 1 c: (d: 1 e: 2))",  # Dictionary containing dictionary
        "(7:a b c d)",              # Canonical S-Expression list with symbol
        "(3:a b 3:c d)"             # Canonical S-Expression list of symbols
    ]

    for payload_in in payloads:
        command, parameters = parse(payload_in)
        print(f"{payload_in} --> command: {command}, parameters: {parameters}")
        payload_out = generate(command, parameters)
        print(f"{command}, {parameters} --> payload: {payload_out}\n")

if __name__ == "__main__":
    main()



# Description
# ~~~~~~~~~~~
# Provides the AikoServices framework for an application process.
# Supports none, one or many Services within a single Process.
#
# Environment variables
# ~~~~~~~~~~~~~~~~~~~~~
# Fine-grained logging control ...
# - Show low-level incoming MQTT messages
#   - AIKO_LOG_LEVEL_MESSAGE: ERROR, WARNING, INFO, DEBUG
#
# - Show Connection locking, MQTT issues, Registrar issues, Services locking
#   - AIKO_LOG_LEVEL_PROCESS: ERROR, WARNING, INFO, DEBUG
#
# Usage
# ~~~~~
# def time_handler():
#     print("timer_handler()")
#
# from aiko_services.main import *       # Invokes __init__.py
# aiko.process = process_create()        # Invoked from __init__.py
# event.add_timer_handler(handler, 1.0)
# aiko.process.run()                     # Invoked by application code
#
# To Do
# ~~~~~
# * Automatically attempt reconnection with MQTT server after disconnection
#
# * BUG: AikoLogger.logger()" uses a Service Id of 0, not actual Service Id
#
# * Implement multiple message server connections for Federation, etc
#
# - A Process with ...
#   - Single Service:    Typically isn't a LifeCycleManager
#   - Multiple Services: First Service is a LifeCycleManager or Pipeline
#     - Use optimized single Process LifeCycleManager (not distributed) ?
#     - Using local or remote functional calls should be very similar ?
#
# - Ensure that add_service() and remove_service() update the Registrar
# - Consider moving all registrar related functions into the Registar
#
# - If a Service event (message, timer) handler raises an Exception,
#   then provide the option to terminate and remove the Service or
#   terminate the entire Process
#
# - Replace global Event functions with a Handler class instance
#
# - Improve AikoLogger (console and MQTT) to accurately provide Service id
#   - Use ContextManager ?
#
# - Review "event.py" message and mailbox queues for latency and bandwidth
#   - Minimal latency from MQTT on_message() handler --> application handler
#   - Minimal latency for priority message handling, whilst handling others
# - Load test 1 Process containing 1,000+ Services
# - Load test 1 Process containing 10,000+ Services (what is the limit ?)
# - Load test 1,000+ Process containing 1,000+ Services
# - Load test 1,000+ Process containing 10,000+ Services (what is the limit ?)
#
# - HL Fleet
#   - HL Serving instances: Provide Ray head and worker node public IP addresses
#   - Registrar History and Recorder --> Dashboard, local and remote storage ?
#   - HostService: monitor CPU, memory, disk, GPU, network, mosquitto, etc
#     - Run on artemis, zeus, AWS EC2 Ray head and worker nodes
#   - Statistics: InfluxDB --> Grafana

import os
import sys
import traceback

from aiko_services.main import *
from aiko_services.main.message import *
from aiko_services.main.utilities import *

__all__ = ["aiko", "process_create"]

_VERSION = 0

# Environment variable names
_AIKO_LOG_LEVEL_MESSAGE = "AIKO_LOG_LEVEL_MESSAGE"
_AIKO_LOG_LEVEL_PROCESS = "AIKO_LOG_LEVEL_PROCESS"

# --------------------------------------------------------------------------- #
# ProcessData is intended as a singleton class data structure only

class ProcessData:
    TOPIC_REGISTRAR_BOOT = f"{get_namespace()}/service/registrar"

    connection = Connection("process")
#   event = aiko_services.event  # TODO: Replace with Handler() instance
    logger = None
    message = None
    process = None
    registrar = None

    topic_path_process = f"{get_namespace()}/{get_hostname()}/{get_pid()}"
    topic_path = f"{topic_path_process}/0"
    topic_in = f"{topic_path}/in"
    topic_log = f"{topic_path}/log"
    topic_lwt = f"{topic_path}/state"
    topic_out = f"{topic_path}/out"
    payload_lwt = "(absent)"

    @classmethod
    def get_topic_path(cls, service_id):
        return f"{cls.topic_path_process}/{service_id}"

aiko = ProcessData

# --------------------------------------------------------------------------- #
# AikoLogger should be usable even before ProcessImplementation() is created

class AikoLogger:
    @classmethod
    def logger(cls,
        name, log_level=None, logging_handler=None, topic=aiko.topic_log):

        if logging_handler is None:
            option = os.environ.get("AIKO_LOG_MQTT", "all")
            if option in ("all", "true"):
                logging_handler = LoggingHandlerMQTT(aiko, topic, option)
        return get_logger(name, log_level, logging_handler)

aiko.logger = AikoLogger.logger

_LOGGER_MESSAGE = aiko.logger(
    f"{__name__}.message",
    log_level=os.environ.get(_AIKO_LOG_LEVEL_MESSAGE, "INFO")
)

_LOGGER_PROCESS = aiko.logger(
    f"{__name__}.process",
    log_level=os.environ.get(_AIKO_LOG_LEVEL_PROCESS, "INFO")
)

# --------------------------------------------------------------------------- #
# ProcessImplementation is intended as a singleton class implementation only

class ProcessImplementation(ProcessData):
    def __init__(self):
        self.initialized = False
        self.running = False  # TODO: Replace with StateMachine (see actor.py)
        self.service_count = 0

        self._exit_status = 0
        self._message_handlers = {}
        self._message_handlers_binary_topics = {}
        self._message_handlers_wildcard_topics = []
        self._registrar_absent_terminate = False
        self._services = {}
        self._services_lock = Lock(f"{__name__}._services", _LOGGER_PROCESS)

    def initialize(self, mqtt_connection_required=True):
        if not self.initialized:
            self.initialized = True
            event.add_queue_handler(self.on_message_queue_handler, ["message"])
            self.add_message_handler(
                self.on_registrar, aiko.TOPIC_REGISTRAR_BOOT)

            aiko.message = Castaway()  # standalone and isolated :(
            mqtt_connected = False
            try:
                aiko.message = MQTT(
                    self.on_message, self._message_handlers,
                    aiko.topic_lwt, aiko.payload_lwt, False,
                    self.on_mqtt_state_change
                )
                mqtt_connected = True
            except SystemError as system_error:
                if mqtt_connection_required:
                    _LOGGER_PROCESS.error(system_error)
                else:
                    _LOGGER_PROCESS.warning(system_error)
            if mqtt_connection_required and not mqtt_connected:
                raise SystemExit()

            context = ContextManager(aiko, aiko.message)

    def run(self, loop_when_no_handlers=False, mqtt_connection_required=True):
        self.initialize(mqtt_connection_required=mqtt_connection_required)

        if not self.running:
            try:
                self.running = True
                event.loop(loop_when_no_handlers)  ## Blocking call ##
            finally:
                self.running = False

        if self._exit_status:
            sys.exit(self._exit_status)

    def add_message_handler(self, message_handler, topic, binary=False):
        if not topic in self._message_handlers:
            self._message_handlers[topic] = []
            if binary:
                self._message_handlers_binary_topics[topic] = True
            if ("#" in topic) or ("+" in topic):
                self._message_handlers_wildcard_topics.append(topic)
            if aiko.message:
                aiko.message.subscribe(topic)
        self._message_handlers[topic].append(message_handler)

    def remove_message_handler(self, message_handler, topic):
        if topic in self._message_handlers:
            if message_handler in self._message_handlers[topic]:
                self._message_handlers[topic].remove(message_handler)
            if len(self._message_handlers[topic]) == 0:
                del self._message_handlers[topic]
                if topic in self._message_handlers_binary_topics:
                    del self._message_handlers_wildcard_topics[topic]
                if topic in self._message_handlers_wildcard_topics:
                    del self._message_handlers_wildcard_topics[topic]
                if self.message:
                    self.message.unsubscribe(topic)

    def _add_service_to_registrar(self, service):
        if service.protocol:
            try:
                owner = get_username()
            except:
                owner = "????????"
                _LOGGER_PROCESS.warning(
                    "Unable to acquire username to identify the Service owner")
            tags = service.get_tags_string()
            # TODO: For payload_out, use parser.generate() ?
            payload_out = f"(add {service.topic_path} {service.name} "  \
                  f"{service.protocol} {service.transport} {owner} ({tags}))"
            registrar_topic_in = f"{aiko.registrar['topic_path']}/in"
            aiko.message.publish(registrar_topic_in, payload_out)

    def _remove_service_from_registrar(self, service):
        if service.protocol:
            # TODO: For payload_out, use parser.generate() ?
            payload_out = f"(remove {service.topic_path})"
            registrar_topic_in = f"{aiko.registrar['topic_path']}/in"
            aiko.message.publish(registrar_topic_in, payload_out)

    def add_service(self, service):
        try:
            self._services_lock.acquire("add_service()")
            self.service_count += 1
            service.service_id = self.service_count
            service.topic_path = aiko.get_topic_path(service.service_id)
            self._services[service.service_id] = service
        finally:
            self._services_lock.release()

        if self.connection.is_connected(ConnectionState.REGISTRAR):
            self._add_service_to_registrar(service)
        return self.service_count

    def remove_service(self, service_id):
        try:
            self._services_lock.acquire("remove_service")
            if service_id in self._services:
                del self._services[service_id]
                self.service_count -= 1
        finally:
            self._services_lock.release()

        if self.connection.is_connected(ConnectionState.REGISTRAR):
            self._remove_service_from_registrar(service)
        return self.service_count

    def on_message(self, mqtt_client, userdata, message):
        try:
            event.queue_put(message, "message")
        except Exception as exception:
            print(traceback.format_exc())

    def on_message_queue_handler(self, message, _):
        topic = message.topic
        payload_in = message.payload
        if topic not in self._message_handlers_binary_topics:
            payload_in = payload_in.decode("utf-8")
        if _LOGGER_MESSAGE.isEnabledFor(DEBUG):  # Don't expand debug message
            _LOGGER_MESSAGE.debug(f"Message: {topic}: {payload_in}")

        message_handler_list = []
        topics_matched = self.topic_matcher(topic, self._message_handlers)
        for topic_match in topics_matched:
            message_handler_list.extend(self._message_handlers[topic_match])

        if len(message_handler_list) > 0:
            for message_handler in message_handler_list:
                try:
                    if message_handler(aiko, topic, payload_in):
                        return
                except Exception as exception:  # REVIEW
                    payload_out = traceback.format_exc()
                    print(payload_out)
                    aiko.message.publish(aiko.topic_log, payload_out)

    def on_mqtt_state_change(self, mqtt_state):
        try:
            self.connection.lock_acquire("on_mqtt_state()")
            connection_state = ConnectionState.NONE
            if mqtt_state == MessageState.CONNECTED:
                connection_state = ConnectionState.TRANSPORT
            self.connection.update_state(connection_state, _LOGGER_PROCESS)
        finally:
            self.connection.lock_release()

    def on_registrar(self, _, topic, payload_in):
        action = None
        parse_okay = False
        registrar = {}
        try:
            command, parameters = parse(payload_in)
            if len(parameters) > 0:
                action = parameters[0]
                if command == "primary":
                    if len(parameters) == 4 and action == "found":
                        registrar["topic_path"] = parameters[1]
                        registrar["version"] = parameters[2]
                        registrar["timestamp"] = parameters[3]
                        parse_okay = True
                    if len(parameters) == 1 and action == "absent":
                        parse_okay = True
            if parse_okay:
                if action == "found":
                    aiko.registrar = registrar
                    try:
                        self.connection.lock_acquire("on_registrar(): found")
                        if self.connection.get_state() != ConnectionState.TRANSPORT:
                            _LOGGER_PROCESS.warning(
                                "Registrar discovered when not in TRANSPORT state")
                        self.connection.update_state(
                            ConnectionState.REGISTRAR, _LOGGER_PROCESS)
                    finally:
                        self.connection.lock_release()

                    try:
                        self._services_lock.acquire("on_registrar() #1")
                        for service in self._services.values():
                            self._add_service_to_registrar(service)
                    finally:
                        self._services_lock.release()

                if action == "absent":
                    aiko.registrar = None
                    try:
                        self.connection.lock_acquire("on_registrar(): absent")
                        if self.connection.get_state() == ConnectionState.REGISTRAR:
                            self.connection.update_state(
                                ConnectionState.TRANSPORT, _LOGGER_PROCESS)
                        else:
                            _LOGGER_PROCESS.warning(
                                "Registrar lost when not in REGISTRAR state")
                        if self._registrar_absent_terminate:
                            self.terminate(1)
                    finally:
                        self.connection.lock_release()

                try:
                    self._services_lock.acquire("on_registrar() #2")
                    for service in self._services.values():
                        service.registrar_handler_call(action, aiko.registrar)
                finally:
                    self._services_lock.release()
        except Exception as exception:
            _LOGGER_PROCESS.warning(
                f"Exception raised when adding to Registrar")
            #   f"Exception raised when adding to Registrar: {exception}\n"
            #   f"{exception.__traceback__.tb_frame} "
            #   f"raised via line {exception.__traceback__.tb_lineno}")
            traceback.print_exception(
                type(exception), exception, exception.__traceback__)

    def set_last_will_and_testament(self,
        topic_lwt, payload_lwt="(absent)", retain_lwt=False):

        aiko.message.set_last_will_and_testament(
            topic_lwt, payload_lwt, retain_lwt)

    def set_registrar_absent_terminate(self):
        self._registrar_absent_terminate = True

    def terminate(self, exit_status=0):
        self._exit_status = exit_status
        event.terminate()

# topic_matcher() matches a ...
# - literal topic without wildcards
# - topic ending with a "#" wildcard, e.g prefix/#
# - topic containing "+" wildcards, e.g prefix/+/+/+/suffix

    def topic_matcher(self, topic, topics):
        if topic in topics:
            topics_matched = [topic]
        else:
            topics_matched = []

        for wildcard_topic in self._message_handlers_wildcard_topics:
            tokens = topic.split("/")
            wildcard_tokens = wildcard_topic.split("/")

            if wildcard_tokens[-1] == "#":
                if tokens[:-1] == wildcard_tokens[:-1]:
                    topics_matched.append(wildcard_topic)
            elif tokens[0] == wildcard_tokens[0] and  \
                 tokens[-1] == wildcard_tokens[-1]:
                    topics_matched.append(wildcard_topic)
        return topics_matched

def process_create():
    if not ProcessData.process:
        ProcessData.process = ProcessImplementation()
    return ProcessData.process

# --------------------------------------------------------------------------- #



# Declaration order is determined by the dependency on static references
#
# To Do
# ~~~~~
# - None, yet !

SERVICE_PROTOCOL_AIKO = "github.com/geekscape/aiko_services/protocol"

REGISTRAR_SERVICE_TYPE = "registrar"
REGISTRAR_VERSION = 2
REGISTRAR_PROTOCOL =  \
    f"{SERVICE_PROTOCOL_AIKO}/{REGISTRAR_SERVICE_TYPE}:{REGISTRAR_VERSION}"

from .context import (
    Context, Interface, ServiceProtocolInterface, ContextService,
    ContextPipelineElement, ContextPipeline,
    service_args, actor_args, pipeline_element_args, pipeline_args
)

from .component import compose_class, compose_instance

from .hook import DEFAULT_HOOK, Hook, Hooks

from .connection import ConnectionState, Connection

from .event import (
    add_flatout_handler, add_mailbox_handler,
    add_queue_handler, add_timer_handler,
    loop, mailbox_put, queue_put,
    remove_flatout_handler, remove_mailbox_handler,
    remove_queue_handler, remove_timer_handler,
    terminate
)

from .process import aiko, process_create

from .lease import Lease

from .service import (
    ServiceFields, ServiceFilter, ServiceProtocol,
    ServiceTags, ServiceTopicPath, Services,
    Service, ServiceImpl
)

from .state import StateMachineOld

from .proxy import ProxyAllMethods, is_callable, proxy_trace

from .share import (
    ECConsumer, PROTOCOL_EC_CONSUMER,
    ECProducer, PROTOCOL_EC_PRODUCER,
    services_cache_create_singleton, services_cache_delete
)

from .actor import (
    Actor, ActorImpl, ActorTest, ActorTestImpl, ActorTopic,
    ACTOR_HOOK_MESSAGE_CALL, ACTOR_HOOK_MESSAGE_IN
)

from .discovery import (
    ServiceDiscovery, ActorDiscovery,
    PipelineElementDiscovery, PipelineDiscovery,
    do_command, do_discovery, do_request, get_service_proxy
)

from .process_manager import ProcessManager

from .lifecycle import LifeCycleClient, LifeCycleManager

from .stream import (
    DEFAULT_STREAM_ID, FIRST_FRAME_ID, Frame, Stream,
    StreamEvent, StreamEventName, StreamState, StreamStateName
)

from .pipeline import (
    Pipeline, PipelineElement, PipelineElementImpl,
    PipelineElementLoop,
    PipelineImpl,
    PIPELINE_HOOK_PROCESS_ELEMENT, PIPELINE_HOOK_PROCESS_ELEMENT_POST,
    PIPELINE_HOOK_PROCESS_FRAME, PROTOCOL_PIPELINE
)

from .scheme import DataScheme

from .source_target import DataSource, DataTarget

from .registrar import *

# from .cli import *

from .dashboard import LogUI, ServiceFrame

aiko.process = process_create()



ENTITY_AVRO_SCHEMA = [
    {
        "name": "Point",
        "namespace": "tracking",
        "type": "record",
        "fields": [{"name": "x", "type": "double"}, {"name": "y", "type": "double"}],
    },
    {
        "name": "Bounds",
        "namespace": "tracking",
        "type": "record",
        "fields": [{"name": "min", "type": "Point"}, {"name": "max", "type": "Point"}],
    },
    {
        "name": "Detection",
        "namespace": "tracking",
        "type": "record",
        "fields": [
            {"name": "frame_id", "type": "int"},
            {"name": "bounds", "type": "Bounds"},
            {"name": "geometry_type", "type": "int"},
            {"name": "confidence", "type": "float"},
        ],
    },
    {
        "name": "EntityDatumSource",
        "namespace": "tracking",
        "type": "record",
        "fields": [{"name": "confidence", "type": "double"}, {"name": "frameId", "type": "int"}],
    },
    {
        "name": "Eavt",
        "namespace": "tracking",
        "type": "record",
        "fields": [
            {"name": "entityId", "type": "string"},
            {"name": "entityAttributeId", "type": "string"},
            {"name": "entityAttributeEnumId", "type": "string"},
            {"name": "entityDatumSource", "type": "EntityDatumSource"},
            {"name": "value", "type": ["string", "int", "long", "float", "double", "null"]},
            {"name": "time", "type": {"type": "int", "logicalType": "date"}},
        ],
    },
    {
        "name": "Track",
        "namespace": "tracking",
        "type": "record",
        "fields": [
            {"name": "track_id", "type": "string"},
            {"name": "detections", "type": {"type": "array", "items": "Detection"}},
            {"name": "eavts", "type": {"type": "array", "items": "Eavt"}},
        ],
    },
    {
        "name": "Entity",
        "namespace": "tracking",
        "type": "record",
        "fields": [
            {"name": "id", "type": "string"},
            {"name": "object_class", "type": "string"},
            {"name": "tracks", "type": {"type": "array", "items": "Track"}},
            {"name": "embeddings", "type": {"type": "array", "items": {"type": "array", "items": "int"}}},
        ],
    },
]



from datetime import datetime, timezone
from typing import Annotated, Any, Dict, List, Optional
from uuid import UUID, uuid4

from pydantic import ConfigDict, Field
from shapely.geometry import Polygon

from highlighter.client.base_models.annotation import Annotation
from highlighter.client.base_models.datum_source import DatumSource
from highlighter.client.base_models.entity import Entity
from highlighter.client.base_models.image import Image
from highlighter.client.base_models.observation import Observation
from highlighter.client.base_models.user import User
from highlighter.core.hl_base_model import HLDataModel, HLModelMap
from highlighter.core.labeled_uuid import LabeledUUID

__all__ = [
    "NEW_Submission",
]


class NEW_Submission(HLDataModel):

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    id: UUID
    data_file_id: int
    entities: HLModelMap = Field(..., default_factory=lambda: HLModelMap(Entity))
    created_at: datetime = Field(..., default_factory=lambda: datetime.now(timezone.utc))

    def to_json(self):
        data = self.model_dump()
        data["id"] = str(data["id"])
        data["created_at"] = data["created_at"].isoformat()
        data["entities"] = [en.to_json() for en in data["entities"].values()]
        return data

    @classmethod
    def from_json(cls, json_path):
        # ToDo
        pass


if __name__ == "__main__":
    import json

    i = 0
    su_id = UUID(int=i)
    su = NEW_Submission(
        id=su_id,
        data_file_id=i,
    )
    for e in range(2):
        en_id = UUID(int=int(f"{i}{e}"))
        en = Entity(id=en_id)
        su.entities.add(en)

        # Add a single global observation
        gl_ob_id = UUID(int=e)
        gl_ob = Observation(
            id=gl_ob_id,
            entity_id=en_id,
            attribute_id=uuid4(),
            value=f"gl_ob_value_{e}",
            datum_source=DatumSource(confidence=0.8),
        )
        en.global_observations.add(gl_ob)

        for a in range(2):
            an_id = UUID(int=int(f"{i}{e}{a}"))
            coords = ((0, 0), (a + 10, 0), (a + 10, a + 10), (0, a + 10), (0, 0))
            an = Annotation(
                id=an_id,
                entity_id=en_id,
                location=Polygon(coords),
                datum_source=DatumSource(confidence=1.0),
            )
            en.annotations.add(an)

            for o in range(2):
                ob_id = UUID(int=int(f"{i}{e}{a}{o}"))

                ob = Observation(
                    id=ob_id,
                    entity_id=en_id,
                    attribute_id=uuid4(),
                    value=f"foo_{i}{e}{a}{o}",
                    datum_source=DatumSource(confidence=0.5),
                )
                an.observations.add(ob)

    with open("test_submission.json", "w") as f:
        json.dump(su.to_json(), f, indent=2)



from pathlib import Path
from typing import Any, Dict, Iterable, Optional, Tuple, Union
from uuid import UUID

import aiko_services as aiko
from aiko_services import PROTOCOL_PIPELINE, ActorTopic
from aiko_services import DataSource as AikoDataSource
from aiko_services import (
    PipelineImpl,
    StreamEvent,
    StreamState,
    compose_instance,
    pipeline_args,
    pipeline_element_args,
)
from pydantic import BaseModel

__all__ = [
    "ActorTopic",
    "Capability",
    "DataSourceCapability",
    "ContextPipelineElement",
    "EntityUUID",
    "PROTOCOL_PIPELINE",
    "PipelineElement",
    "PipelineImpl",
    "StreamEvent",
    "StreamState",
    "compose_instance",
    "compose_instance",
    "pipeline_args",
    "pipeline_element_args",
]

EntityUUID = UUID

"""Decouple the rest of the code from aiko.PipelineElement"""
ContextPipelineElement = aiko.ContextPipelineElement
PipelineElement = aiko.PipelineElement

# SEPARATOR = b"\x1c"  # ASCII 28 (File Separator)
SEPARATOR = 28  # ASCII 28 (File Separator)


class _BaseCapability:
    class DefaultStreamParameters(BaseModel):
        """Populate with default stream param key fields"""

        pass

    @classmethod
    def default_stream_parameters(cls) -> BaseModel:
        return {
            k: v.default for k, v in cls.DefaultStreamParameters.model_fields.items() if not v.is_required()
        }

    def _get_parameter(
        self, name, default=None, required=False, use_pipeline=True, self_share_priority=True
    ) -> Tuple[Any, bool]:
        """Adds the correct output type to get_parameter type checking
        does not complain
        """
        return self.get_parameter(
            name,
            default=default,
            required=required,
            use_pipeline=use_pipeline,
            self_share_priority=self_share_priority,
        )


class Capability(PipelineElement, _BaseCapability):

    def __init__(self, context: aiko.ContextPipelineElement):
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, *args) -> Tuple[StreamEvent, dict]:
        raise NotImplementedError()

    def start_stream(self, stream, stream_id, use_create_frame=True):
        validated_parameters = self.DefaultStreamParameters(**self.parameters)
        for param_name in self.DefaultStreamParameters.model_fields:
            self.parameters[f"{self.definition.name}.{param_name}"] = getattr(
                validated_parameters, param_name
            )
            self.parameters.pop(param_name, None)
        stream.parameters.update(self.parameters)
        return StreamEvent.OKAY, {}


# ToDO: Remove
class DataSourceType(BaseModel):
    # class MediaType(str, Enum):
    #    IMAGE = "IMAGE"
    #    TEXT = "TEXT"
    #    VIDEO = "VIDEO"

    media_type: str
    url: str
    id: UUID
    content: Optional[Any] = None

    @classmethod
    def image_iter(cls, images: Iterable[Union[str, Path, bytes]]):
        pass

    @classmethod
    def video_iter(cls, videos: Iterable[Union[str, Path, bytes]]):
        pass

    @classmethod
    def text_iter(cls, tests: Iterable[Union[str, Path, bytes]]):
        pass


class DataSourceCapability(AikoDataSource, _BaseCapability):

    stream_media_type = None

    class DefaultStreamParameters(_BaseCapability.DefaultStreamParameters):

        rate: Optional[float] = None
        batch_size: int = 1
        data_sources: Optional[str] = None
        file_ids: Optional[Iterable] = None
        task_id: Optional[UUID] = None

    @property
    def rate(self) -> float:
        return self._get_parameter("rate")[0]

    @property
    def batch_size(self) -> int:
        return self._get_parameter("batch_size")[0]

    def __init__(self, context: aiko.ContextPipelineElement):
        context.get_implementation("PipelineElement").__init__(self, context)

    def frame_generator(self, stream, pipeline_iter_idx):
        """Produce a batch of frames.

        Args:
            stream: The Stream context
            pipeline_iter_idx: An integer counting the number of times the
                               pipeline has been executed, (ie: process_frame
                               has been called)

        """
        batch_size = self.batch_size
        task_id, _ = self._get_parameter("task_id")

        frame_data_batch = {"data_files": [], "entities": {}}
        for _ in range(batch_size):
            try:
                data_file, entities = self.get_next_frame_data(stream)
                frame_data_batch["data_files"].append(data_file)
                frame_data_batch["entities"].update(entities)
                self.logger.debug(f"data_file: {data_file}, entities: {entities}")
            except StopIteration:
                pass
            except Exception as e:
                return StreamEvent.ERROR, {"diagnostic": e}

        if not frame_data_batch["data_files"]:
            return StreamEvent.STOP, {"diagnostic": "All frames generated"}

        # For each pipeline iteration the is a batch of file_ids and frame_ids
        stream.variables["task_id"] = task_id

        return StreamEvent.OKAY, frame_data_batch

    def start_stream(self, stream, stream_id):
        stream.variables["video_capture"] = None
        stream.variables["video_frame_generator"] = None

        return super().start_stream(
            stream, stream_id, frame_generator=self.frame_generator, use_create_frame=False
        )

    def get_next_frame_data(self, stream):
        raise NotImplementedError()

    def process_frame(self, stream, data_files, entities) -> Tuple[StreamEvent, Dict]:
        return StreamEvent.OKAY, {"data_files": data_files, "entities": entities}

    def using_hl_data_scheme(self, stream) -> bool:
        return "hl_source_data" in stream.variables
